@@ -1,4 +1,4 @@
-// Copyright (c) 2022-2023 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2022-2024 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -41,6 +41,7 @@
 #include <chrono>
 #include <iostream>
 #include <iterator>
+#include <type_traits>
 
 #include <cstddef>
 
@@ -110,9 +111,12 @@ hipError_t adjacent_difference_impl(void* const          temporary_storage,
                                     const hipStream_t    stream,
                                     const bool           debug_synchronous)
 {
-    using value_type = typename std::iterator_traits<InputIt>::value_type;
+    using value_type  = typename std::iterator_traits<InputIt>::value_type;
+    using output_type = rocprim::invoke_result_binary_op_t<value_type, BinaryFunction>;
+    using larger_type
+        = std::conditional_t<(sizeof(value_type) >= sizeof(output_type)), value_type, output_type>;
 
-    using config = wrapped_adjacent_difference_config<Config, InPlace, value_type>;
+    using config = wrapped_adjacent_difference_config<Config, InPlace, larger_type>;
 
     detail::target_arch target_arch;
     hipError_t          result = detail::host_target_arch(stream, target_arch);
@@ -154,9 +158,10 @@ hipError_t adjacent_difference_impl(void* const          temporary_storage,
         // next block, otherwise the first item is needed for the previous block
         const auto offset = items_per_block - (Right ? 0 : 1);
 
-        const auto block_starts_iter = make_transform_iterator(
-            rocprim::make_counting_iterator(std::size_t{0}),
-            [=, base = input + offset](std::size_t i) { return base[i * items_per_block]; });
+        const auto block_starts_iter
+            = make_transform_iterator(rocprim::make_counting_iterator(std::size_t{0}),
+                                      [=, base = input + offset](std::size_t i) -> value_type
+                                      { return base[i * items_per_block]; });
 
         const hipError_t error = ::rocprim::transform(block_starts_iter,
                                                       previous_values,
@@ -244,26 +249,25 @@ hipError_t adjacent_difference_impl(void* const          temporary_storage,
 /// }
 /// \endcode
 ///
-/// \tparam Config - [optional] configuration of the primitive. It has to be
-/// `adjacent_difference_config` or a class derived from it.
-/// \tparam InputIt - [inferred] random-access iterator type of the input range. Must meet the
+/// \tparam Config [optional] configuration of the primitive, must be `default_config` or `adjacent_difference_config`.
+/// \tparam InputIt [inferred] random-access iterator type of the input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam OutputIt - [inferred] random-access iterator type of the output range. Must meet the
+/// \tparam OutputIt [inferred] random-access iterator type of the output range. Must meet the
 /// requirements of a C++ OutputIterator concept. It can be a simple pointer type.
-/// \tparam BinaryFunction - [inferred] binary operation function object that will be applied to
+/// \tparam BinaryFunction [inferred] binary operation function object that will be applied to
 /// consecutive items. The signature of the function should be equivalent to the following:
 /// `U f(const T1& a, const T2& b)`. The signature does not need to have
 /// `const &`, but function object must not modify the object passed to it
-/// \param temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// `storage_size` and function returns without performing the scan operation
-/// \param storage_size - reference to a size (in bytes) of `temporary_storage`
-/// \param input - iterator to the input range
-/// \param output - iterator to the output range, must have any overlap with input
-/// \param size - number of items in the input
-/// \param op - [optional] the binary operation to apply
-/// \param stream - [optional] HIP stream object. Default is `0` (the default stream)
-/// \param debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param storage_size reference to a size (in bytes) of `temporary_storage`
+/// \param input iterator to the input range
+/// \param output iterator to the output range, must not have any overlap with input.
+/// \param size number of items in the input
+/// \param op [optional] the binary operation to apply
+/// \param stream [optional] HIP stream object. Default is `0` (the default stream)
+/// \param debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors and extra debugging info is printed to the
 /// standard output. Default value is `false`
 ///
@@ -340,23 +344,22 @@ hipError_t adjacent_difference(void* const          temporary_storage,
 /// }
 /// \endcode
 ///
-/// \tparam Config - [optional] configuration of the primitive. It has to be
-/// `adjacent_difference_config` or a class derived from it.
-/// \tparam InputIt - [inferred] random-access iterator type of the value range. Must meet the
+/// \tparam Config [optional] configuration of the primitive, must be `default_config` or `adjacent_difference_config`.
+/// \tparam InputIt [inferred] random-access iterator type of the value range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam BinaryFunction - [inferred] binary operation function object that will be applied to
+/// \tparam BinaryFunction [inferred] binary operation function object that will be applied to
 /// consecutive items. The signature of the function should be equivalent to the following:
 /// `U f(const T1& a, const T2& b)`. The signature does not need to have
 /// `const &`, but function object must not modify the object passed to it
-/// \param temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// `storage_size` and function returns without performing the scan operation
-/// \param storage_size - reference to a size (in bytes) of `temporary_storage`
-/// \param values - iterator to the range values, will be overwritten with the results
-/// \param size - number of items in the input
-/// \param op - [optional] the binary operation to apply
-/// \param stream - [optional] HIP stream object. Default is `0` (the default stream)
-/// \param debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param storage_size reference to a size (in bytes) of `temporary_storage`
+/// \param values iterator to the range values, will be overwritten with the results
+/// \param size number of items in the input
+/// \param op [optional] the binary operation to apply
+/// \param stream [optional] HIP stream object. Default is `0` (the default stream)
+/// \param debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors and extra debugging info is printed to the
 /// standard output. Default value is `false`
 ///
@@ -379,6 +382,64 @@ hipError_t adjacent_difference_inplace(void* const          temporary_storage,
         temporary_storage, storage_size, values, values, size, op, stream, debug_synchronous);
 }
 
+/// \brief Parallel primitive for applying a binary operation across pairs of consecutive elements
+/// in device accessible memory. Writes the output to the position of the left item.
+///
+/// \tparam Config [optional] configuration of the primitive, must be `default_config` or `adjacent_difference_config`.
+/// \tparam InputIt [inferred] random-access iterator type of the value range. Must meet the
+/// requirements of a C++ InputIterator concept. It can be a simple pointer type.
+/// \tparam OutputIt [inferred] random-access iterator type of the output range. Must meet the
+/// requirements of a C++ OutputIterator concept. It can be a simple pointer type.
+/// \tparam BinaryFunction [inferred] binary operation function object that will be applied to
+/// consecutive items. The signature of the function should be equivalent to the following:
+/// `U f(const T1& a, const T2& b)`. The signature does not need to have
+/// `const &`, but function object must not modify the object passed to it
+///
+/// \param temporary_storage pointer to a device-accessible temporary storage. When
+/// a null pointer is passed, the required allocation size (in bytes) is written to
+/// `storage_size` and function returns without performing the scan operation
+/// \param storage_size reference to a size (in bytes) of `temporary_storage`
+/// \param input iterator to the range values
+/// \param output iterator to the output range. Allowed to point to the same elements as `input`.
+///   Only complete overlap or no overlap at all is allowed between `input` and `output`. In other words
+///   writing to `output[i]` is only allowed to overwrite `input[i]`, any other element must not be changed.
+/// \param size number of items in the input
+/// \param op [optional] the binary operation to apply
+/// \param stream [optional] HIP stream object. Default is `0` (the default stream)
+/// \param debug_synchronous [optional] If true, synchronization after every kernel
+/// launch is forced in order to check for errors and extra debugging info is printed to the
+/// standard output. Default value is `false`
+///
+/// \return `hipSuccess` (0) on success, otherwise the HIP runtime error of
+/// type `hipError_t`
+///
+/// \note This function has to perform an extra copy due to (potentially) writing its values in-place. If it is known that `input` and `output`
+/// don't overlap then adjacent_difference should be preferred as it avoids this extra copy.
+template<typename Config = default_config,
+         typename InputIt,
+         typename OutputIt,
+         typename BinaryFunction = ::rocprim::minus<>>
+hipError_t adjacent_difference_inplace(void* const          temporary_storage,
+                                       std::size_t&         storage_size,
+                                       const InputIt        input,
+                                       const OutputIt       output,
+                                       const std::size_t    size,
+                                       const BinaryFunction op                = BinaryFunction{},
+                                       const hipStream_t    stream            = 0,
+                                       const bool           debug_synchronous = false)
+{
+    static constexpr bool in_place = true;
+    static constexpr bool right    = false;
+    return detail::adjacent_difference_impl<Config, in_place, right>(temporary_storage,
+                                                                     storage_size,
+                                                                     input,
+                                                                     output,
+                                                                     size,
+                                                                     op,
+                                                                     stream,
+                                                                     debug_synchronous);
+}
+
 /// \brief Parallel primitive for applying a binary operation across pairs of consecutive elements
 /// in device accessible memory. Writes the output to the position of the right item.
 ///
@@ -393,26 +454,25 @@ hipError_t adjacent_difference_inplace(void* const          temporary_storage,
 /// }
 /// \endcode
 ///
-/// \tparam Config - [optional] configuration of the primitive. It has to be
-/// `adjacent_difference_config` or a class derived from it.
-/// \tparam InputIt - [inferred] random-access iterator type of the input range. Must meet the
+/// \tparam Config [optional] configuration of the primitive, must be `default_config` or `adjacent_difference_config`.
+/// \tparam InputIt [inferred] random-access iterator type of the input range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam OutputIt - [inferred] random-access iterator type of the output range. Must meet the
+/// \tparam OutputIt [inferred] random-access iterator type of the output range. Must meet the
 /// requirements of a C++ OutputIterator concept. It can be a simple pointer type.
-/// \tparam BinaryFunction - [inferred] binary operation function object that will be applied to
+/// \tparam BinaryFunction [inferred] binary operation function object that will be applied to
 /// consecutive items. The signature of the function should be equivalent to the following:
 /// `U f(const T1& a, const T2& b)`. The signature does not need to have
 /// `const &`, but function object must not modify the object passed to it
-/// \param temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// `storage_size` and function returns without performing the scan operation
-/// \param storage_size - reference to a size (in bytes) of `temporary_storage`
-/// \param input - iterator to the input range
-/// \param output - iterator to the output range, must have any overlap with input
-/// \param size - number of items in the input
-/// \param op - [optional] the binary operation to apply
-/// \param stream - [optional] HIP stream object. Default is `0` (the default stream)
-/// \param debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param storage_size reference to a size (in bytes) of `temporary_storage`
+/// \param input iterator to the input range
+/// \param output iterator to the output range, must not have any overlap with input.
+/// \param size number of items in the input
+/// \param op [optional] the binary operation to apply
+/// \param stream [optional] HIP stream object. Default is `0` (the default stream)
+/// \param debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors and extra debugging info is printed to the
 /// standard output. Default value is `false`
 ///
@@ -489,23 +549,22 @@ hipError_t adjacent_difference_right(void* const          temporary_storage,
 /// }
 /// \endcode
 ///
-/// \tparam Config - [optional] configuration of the primitive. It has to be
-/// `adjacent_difference_config` or a class derived from it.
-/// \tparam InputIt - [inferred] random-access iterator type of the value range. Must meet the
+/// \tparam Config [optional] configuration of the primitive, must be `default_config` or `adjacent_difference_config`.
+/// \tparam InputIt [inferred] random-access iterator type of the value range. Must meet the
 /// requirements of a C++ InputIterator concept. It can be a simple pointer type.
-/// \tparam BinaryFunction - [inferred] binary operation function object that will be applied to
+/// \tparam BinaryFunction [inferred] binary operation function object that will be applied to
 /// consecutive items. The signature of the function should be equivalent to the following:
 /// `U f(const T1& a, const T2& b)`. The signature does not need to have
 /// `const &`, but function object must not modify the object passed to it
-/// \param temporary_storage - pointer to a device-accessible temporary storage. When
+/// \param temporary_storage pointer to a device-accessible temporary storage. When
 /// a null pointer is passed, the required allocation size (in bytes) is written to
 /// `storage_size` and function returns without performing the scan operation
-/// \param storage_size - reference to a size (in bytes) of `temporary_storage`
-/// \param values - iterator to the range values, will be overwritten with the results
-/// \param size - number of items in the input
-/// \param op - [optional] the binary operation to apply
-/// \param stream - [optional] HIP stream object. Default is `0` (the default stream)
-/// \param debug_synchronous - [optional] If true, synchronization after every kernel
+/// \param storage_size reference to a size (in bytes) of `temporary_storage`
+/// \param values iterator to the range values, will be overwritten with the results
+/// \param size number of items in the input
+/// \param op [optional] the binary operation to apply
+/// \param stream [optional] HIP stream object. Default is `0` (the default stream)
+/// \param debug_synchronous [optional] If true, synchronization after every kernel
 /// launch is forced in order to check for errors and extra debugging info is printed to the
 /// standard output. Default value is `false`
 ///
@@ -528,6 +587,63 @@ hipError_t adjacent_difference_right_inplace(void* const          temporary_stor
         temporary_storage, storage_size, values, values, size, op, stream, debug_synchronous);
 }
 
+/// \brief Parallel primitive for applying a binary operation across pairs of consecutive elements
+/// in device accessible memory. Writes the output to the position of the right item.
+///
+/// \tparam Config [optional] configuration of the primitive, must be `default_config` or `adjacent_difference_config`.
+/// \tparam InputIt [inferred] random-access iterator type of the value range. Must meet the
+/// requirements of a C++ InputIterator concept. It can be a simple pointer type.
+/// \tparam OutputIt [inferred] random-access iterator type of the output range. Must meet the
+/// requirements of a C++ OutputIterator concept. It can be a simple pointer type.
+/// \tparam BinaryFunction [inferred] binary operation function object that will be applied to
+/// consecutive items. The signature of the function should be equivalent to the following:
+/// `U f(const T1& a, const T2& b)`. The signature does not need to have
+/// `const &`, but function object must not modify the object passed to it
+
+/// \param temporary_storage pointer to a device-accessible temporary storage. When
+/// a null pointer is passed, the required allocation size (in bytes) is written to
+/// `storage_size` and function returns without performing the scan operation
+/// \param storage_size reference to a size (in bytes) of `temporary_storage`
+/// \param input iterator to the range values, will be overwritten with the results
+/// \param output iterator to the output range. Allowed to point to the same elements as `input`.
+///   Only complete overlap or no overlap at all is allowed between `input` and `output`. In other words
+///   writing to `output[i]` is only allowed to overwrite `input[i]`, any other element must not be changed.
+/// \param size number of items in the input
+/// \param op [optional] the binary operation to apply
+/// \param stream [optional] HIP stream object. Default is `0` (the default stream)
+/// \param debug_synchronous [optional] If true, synchronization after every kernel
+/// launch is forced in order to check for errors and extra debugging info is printed to the
+/// standard output. Default value is `false`
+///
+/// \return `hipSuccess` (0) on success, otherwise the HIP runtime error of
+/// type `hipError_t`
+/// \note This function has to perform an extra copy due to (potentially) writing its values in-place. If it is known that `input` and `output`
+/// don't overlap then adjacent_difference_right should be preferred as it avoids this extra copy.
+template<typename Config = default_config,
+         typename InputIt,
+         typename OutputIt,
+         typename BinaryFunction = ::rocprim::minus<>>
+hipError_t adjacent_difference_right_inplace(void* const          temporary_storage,
+                                             std::size_t&         storage_size,
+                                             const InputIt        input,
+                                             const OutputIt       output,
+                                             const std::size_t    size,
+                                             const BinaryFunction op     = BinaryFunction{},
+                                             const hipStream_t    stream = 0,
+                                             const bool           debug_synchronous = false)
+{
+    static constexpr bool in_place = true;
+    static constexpr bool right    = true;
+    return detail::adjacent_difference_impl<Config, in_place, right>(temporary_storage,
+                                                                     storage_size,
+                                                                     input,
+                                                                     output,
+                                                                     size,
+                                                                     op,
+                                                                     stream,
+                                                                     debug_synchronous);
+}
+
 /// @}
 // end of group devicemodule
 
