@@ -1,6 +1,6 @@
 // MIT License
 //
-// Copyright (c) 2017-2022 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -72,6 +72,7 @@ typedef ::testing::Types<
     DeviceMergeParams<unsigned long, unsigned int, rocprim::greater<unsigned long>>,
     DeviceMergeParams<float, custom_double2>,
     DeviceMergeParams<int, float>,
+    DeviceMergeParams<double, double>,
     DeviceMergeParams<int8_t, int8_t>,
     DeviceMergeParams<uint8_t, uint8_t>,
     DeviceMergeParams<rocprim::half, rocprim::half, rocprim::less<rocprim::half>>,
@@ -188,40 +189,35 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKey)
             test_utils::bounds_checking_iterator<key_type> d_keys_checking_output(
                 d_keys_output,
                 out_of_bounds.device_pointer(),
-                size1 + size2
-            );
+                size1 + size2);
 
-            hipGraph_t graph;
-            hipGraphExec_t graph_instance;
-            if (TestFixture::use_graphs)
-                graph = test_utils::createGraphHelper(stream);
-            
             // temp storage
             size_t temp_storage_size_bytes;
             void * d_temp_storage = nullptr;
             // Get size of d_temp_storage
-            HIP_CHECK(
-                rocprim::merge(
-                    d_temp_storage, temp_storage_size_bytes,
-                    d_keys_input1, d_keys_input2,
-                    d_keys_checking_output,
-                    keys_input1.size(), keys_input2.size(),
-                    compare_op, stream, debug_synchronous
-                )
-            );
+            HIP_CHECK(rocprim::merge(d_temp_storage,
+                                     temp_storage_size_bytes,
+                                     d_keys_input1,
+                                     d_keys_input2,
+                                     d_keys_checking_output,
+                                     keys_input1.size(),
+                                     keys_input2.size(),
+                                     compare_op,
+                                     stream,
+                                     debug_synchronous));
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-            
             // temp_storage_size_bytes must be >0
             ASSERT_GT(temp_storage_size_bytes, 0);
 
             // allocate temporary storage
             HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
 
-            if (TestFixture::use_graphs)
-                test_utils::resetGraphHelper(graph, graph_instance, stream);
-            
+            test_utils::GraphHelper gHelper;
+            if(TestFixture::use_graphs)
+            {
+                gHelper.startStreamCapture(stream);
+            }
+
             // Run
             HIP_CHECK(
                 rocprim::merge(
@@ -233,8 +229,10 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKey)
                 )
             );
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
+            if(TestFixture::use_graphs)
+            {
+                gHelper.createAndLaunchGraph(stream);
+            }
 
             HIP_CHECK(hipGetLastError());
             HIP_CHECK(hipDeviceSynchronize());
@@ -259,7 +257,7 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKey)
             hipFree(d_temp_storage);
 
             if (TestFixture::use_graphs)
-                test_utils::cleanupGraphHelper(graph, graph_instance);
+                gHelper.cleanupGraphHelper();
         }
     }
 
@@ -395,14 +393,8 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKeyValue)
             test_utils::bounds_checking_iterator<value_type> d_values_checking_output(
                 d_values_output,
                 out_of_bounds.device_pointer(),
-                size1 + size2
-            );
+                size1 + size2);
 
-            hipGraph_t graph;
-            hipGraphExec_t graph_instance;
-            if (TestFixture::use_graphs)
-                graph = test_utils::createGraphHelper(stream);
-            
             // temp storage
             size_t temp_storage_size_bytes;
             void * d_temp_storage = nullptr;
@@ -419,18 +411,18 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKeyValue)
                 )
             );
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-
             // temp_storage_size_bytes must be >0
             ASSERT_GT(temp_storage_size_bytes, 0);
 
             // allocate temporary storage
             HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
 
-            if (TestFixture::use_graphs)
-                test_utils::resetGraphHelper(graph, graph_instance, stream);
-            
+            test_utils::GraphHelper gHelper;
+            if(TestFixture::use_graphs)
+            {
+                gHelper.startStreamCapture(stream);
+            }
+
             // Run
             HIP_CHECK(
                 rocprim::merge(
@@ -444,9 +436,11 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKeyValue)
                 )
             );
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
-            
+            if(TestFixture::use_graphs)
+            {
+                gHelper.createAndLaunchGraph(stream, true, false);
+            }
+
             HIP_CHECK(hipGetLastError());
             HIP_CHECK(hipDeviceSynchronize());
 
@@ -487,7 +481,7 @@ TYPED_TEST(RocprimDeviceMergeTests, MergeKeyValue)
             hipFree(d_temp_storage);
 
             if (TestFixture::use_graphs)
-                test_utils::cleanupGraphHelper(graph, graph_instance);
+                gHelper.cleanupGraphHelper();
         }
     }
 
@@ -535,13 +529,10 @@ void testMergeMismatchedIteratorTypes()
     static constexpr bool debug_synchronous = false;
 
     hipStream_t stream = 0; // default
-    hipGraph_t graph;
-    hipGraphExec_t graph_instance;
     if (UseGraphs)
     {
         // Default stream does not support hipGraph stream capture, so create one
         HIP_CHECK(hipStreamCreateWithFlags(&stream, hipStreamNonBlocking));
-        graph = test_utils::createGraphHelper(stream);
     }
 
     size_t temp_storage_size_bytes = 0;
@@ -556,16 +547,16 @@ void testMergeMismatchedIteratorTypes()
                              stream,
                              debug_synchronous));
 
-    if (UseGraphs)
-        graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-    
     ASSERT_GT(temp_storage_size_bytes, 0);
 
     void* d_temp_storage = nullptr;
     HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
 
-    if (UseGraphs)
-        test_utils::resetGraphHelper(graph, graph_instance, stream);
+    test_utils::GraphHelper gHelper;
+    if(UseGraphs)
+    {
+        gHelper.startStreamCapture(stream);
+    }
 
     HIP_CHECK(rocprim::merge(d_temp_storage,
                              temp_storage_size_bytes,
@@ -578,8 +569,10 @@ void testMergeMismatchedIteratorTypes()
                              hipStreamDefault,
                              debug_synchronous));
 
-    if (UseGraphs)
-        graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
+    if(UseGraphs)
+    {
+        gHelper.createAndLaunchGraph(stream);
+    }
 
     std::vector<int> keys_output(expected_keys_output.size());
     HIP_CHECK(hipMemcpy(keys_output.data(),
@@ -595,7 +588,7 @@ void testMergeMismatchedIteratorTypes()
 
     if (UseGraphs)
     {
-        test_utils::cleanupGraphHelper(graph, graph_instance);
+        gHelper.cleanupGraphHelper();
         HIP_CHECK(hipStreamDestroy(stream));
     }
 }
