@@ -1,6 +1,6 @@
 // MIT License
 //
-// Copyright (c) 2017-2023 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -32,18 +32,18 @@
 #include "test_utils_types.hpp"
 
 // Params for tests
-template<
-    class InputType,
-    class OutputType = InputType,
-    class FlagType = unsigned int,
-    bool UseIdentityIterator = false,
-    bool UseGraphs = false
->
+template<class InputType,
+         class OutputType         = InputType,
+         class FlagType           = unsigned int,
+         class Config             = rocprim::default_config,
+         bool UseIdentityIterator = false,
+         bool UseGraphs           = false>
 struct DevicePartitionParams
 {
     using input_type = InputType;
     using output_type = OutputType;
     using flag_type = FlagType;
+    using config                                = Config;
     static constexpr bool use_identity_iterator = UseIdentityIterator;
     static constexpr bool use_graphs = UseGraphs;
 };
@@ -55,22 +55,32 @@ public:
     using input_type = typename Params::input_type;
     using output_type = typename Params::output_type;
     using flag_type = typename Params::flag_type;
+    using config                                = typename Params::config;
     const bool debug_synchronous = false;
     static constexpr bool use_identity_iterator = Params::use_identity_iterator;
     static constexpr bool use_graphs = Params::use_graphs;
 };
 
+using config = rocprim::select_config<512,
+                                      1,
+                                      ::rocprim::block_load_method::block_load_transpose,
+                                      ::rocprim::block_load_method::block_load_transpose,
+                                      ::rocprim::block_load_method::block_load_transpose,
+                                      ::rocprim::block_scan_algorithm::using_warp_scan>;
+
 typedef ::testing::Types<
-    DevicePartitionParams<int, int, unsigned char, true>,
+    DevicePartitionParams<int, int, unsigned char, rocprim::default_config, true>,
     DevicePartitionParams<unsigned int, unsigned long>,
     DevicePartitionParams<unsigned char, float>,
+    DevicePartitionParams<float, float, unsigned int, config>,
+    DevicePartitionParams<double, double>,
     DevicePartitionParams<int8_t, int8_t>,
     DevicePartitionParams<uint8_t, uint8_t>,
     DevicePartitionParams<rocprim::half, rocprim::half>,
     DevicePartitionParams<rocprim::bfloat16, rocprim::bfloat16>,
     DevicePartitionParams<test_utils::custom_test_type<long long>>,
-    DevicePartitionParams<int, int, unsigned int, false, true>
-> RocprimDevicePartitionTestsParams;
+    DevicePartitionParams<int, int, unsigned int, rocprim::default_config, false, true>>
+    RocprimDevicePartitionTestsParams;
 
 TYPED_TEST_SUITE(RocprimDevicePartitionTests, RocprimDevicePartitionTestsParams);
 
@@ -83,6 +93,7 @@ TYPED_TEST(RocprimDevicePartitionTests, Flagged)
     using T = typename TestFixture::input_type;
     using U = typename TestFixture::output_type;
     using F = typename TestFixture::flag_type;
+    using config                                = typename TestFixture::config;
     static constexpr bool use_identity_iterator = TestFixture::use_identity_iterator;
     const bool debug_synchronous = TestFixture::debug_synchronous;
 
@@ -137,15 +148,11 @@ TYPED_TEST(RocprimDevicePartitionTests, Flagged)
             }
             std::reverse(expected_rejected.begin(), expected_rejected.end());
 
-            hipGraph_t graph;
-            hipGraphExec_t graph_instance;
-            if (TestFixture::use_graphs)
-                graph = test_utils::createGraphHelper(stream);
-            
             // temp storage
             size_t temp_storage_size_bytes;
             // Get size of d_temp_storage
-            HIP_CHECK(rocprim::partition(
+
+            HIP_CHECK(rocprim::partition<config>(
                 nullptr,
                 temp_storage_size_bytes,
                 d_input,
@@ -156,9 +163,6 @@ TYPED_TEST(RocprimDevicePartitionTests, Flagged)
                 stream,
                 debug_synchronous));
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-            
             // temp_storage_size_bytes must be >0
             ASSERT_GT(temp_storage_size_bytes, 0);
 
@@ -166,11 +170,14 @@ TYPED_TEST(RocprimDevicePartitionTests, Flagged)
             void* d_temp_storage = nullptr;
             HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
 
-            if (TestFixture::use_graphs)
-                test_utils::resetGraphHelper(graph, graph_instance, stream);
-            
+            test_utils::GraphHelper gHelper;;
+            if(TestFixture::use_graphs)
+            {
+                gHelper.startStreamCapture(stream);
+            }
+
             // Run
-            HIP_CHECK(rocprim::partition(
+            HIP_CHECK(rocprim::partition<config>(
                 d_temp_storage,
                 temp_storage_size_bytes,
                 d_input,
@@ -181,9 +188,11 @@ TYPED_TEST(RocprimDevicePartitionTests, Flagged)
                 stream,
                 debug_synchronous));
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-            
+            if(TestFixture::use_graphs)
+            {
+                gHelper.createAndLaunchGraph(stream);
+            }
+
             // Check if number of selected value is as expected_selected
             unsigned int selected_count_output = 0;
             HIP_CHECK(hipMemcpy(&selected_count_output,
@@ -213,14 +222,18 @@ TYPED_TEST(RocprimDevicePartitionTests, Flagged)
             hipFree(d_output);
             hipFree(d_selected_count_output);
             hipFree(d_temp_storage);
-            
-            if (TestFixture::use_graphs)
-                test_utils::cleanupGraphHelper(graph, graph_instance);
+
+            if(TestFixture::use_graphs)
+            {
+                gHelper.cleanupGraphHelper();
+            }
         }
     }
 
-    if (TestFixture::use_graphs)
+    if(TestFixture::use_graphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
 
 TYPED_TEST(RocprimDevicePartitionTests, PredicateEmptyInput)
@@ -231,6 +244,7 @@ TYPED_TEST(RocprimDevicePartitionTests, PredicateEmptyInput)
 
     using T = typename TestFixture::input_type;
     using U = typename TestFixture::output_type;
+    using config                 = typename TestFixture::config;
     const bool debug_synchronous = TestFixture::debug_synchronous;
 
     hipStream_t stream = 0; // default stream
@@ -257,53 +271,49 @@ TYPED_TEST(RocprimDevicePartitionTests, PredicateEmptyInput)
                         hipMemcpyHostToDevice));
 
     test_utils::out_of_bounds_flag out_of_bounds;
-    test_utils::bounds_checking_iterator<U> d_checking_output(
-        d_output,
-        out_of_bounds.device_pointer(),
-        0
-    );
+    test_utils::bounds_checking_iterator<U> d_checking_output(d_output,
+                                                              out_of_bounds.device_pointer(),
+                                                              0);
 
-    hipGraph_t graph;
-    hipGraphExec_t graph_instance;
-    if (TestFixture::use_graphs)
-        graph = test_utils::createGraphHelper(stream);
-    
     // temp storage
     size_t temp_storage_size_bytes;
     // Get size of d_temp_storage
-    HIP_CHECK(rocprim::partition(nullptr,
-                                 temp_storage_size_bytes,
-                                 rocprim::make_constant_iterator<T>(T(345)),
-                                 d_checking_output,
-                                 d_selected_count_output,
-                                 0,
-                                 select_op,
-                                 stream,
-                                 debug_synchronous));
+    HIP_CHECK(rocprim::partition<config>(nullptr,
+                                         temp_storage_size_bytes,
+                                         rocprim::make_constant_iterator<T>(T(345)),
+                                         d_checking_output,
+                                         d_selected_count_output,
+                                         0,
+                                         select_op,
+                                         stream,
+                                         debug_synchronous));
 
-    if (TestFixture::use_graphs)
-        graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-    
     // allocate temporary storage
     void* d_temp_storage = nullptr;
     HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
 
-    if (TestFixture::use_graphs)
-        test_utils::resetGraphHelper(graph, graph_instance, stream);
-    
-    // Run
-    HIP_CHECK(rocprim::partition(d_temp_storage,
-                                 temp_storage_size_bytes,
-                                 rocprim::make_constant_iterator<T>(T(345)),
-                                 d_checking_output,
-                                 d_selected_count_output,
-                                 0,
-                                 select_op,
-                                 stream,
-                                 debug_synchronous));
+    test_utils::GraphHelper gHelper;;
+    if(TestFixture::use_graphs)
+    {
+        gHelper.startStreamCapture(stream);
+    }
 
-    if (TestFixture::use_graphs)
-        graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
+    // Run
+    HIP_CHECK(rocprim::partition<config>(d_temp_storage,
+                                         temp_storage_size_bytes,
+                                         rocprim::make_constant_iterator<T>(T(345)),
+                                         d_checking_output,
+                                         d_selected_count_output,
+                                         0,
+                                         select_op,
+                                         stream,
+                                         debug_synchronous));
+
+    
+    if(TestFixture::use_graphs)
+    {
+        gHelper.createAndLaunchGraph(stream, true, false);
+    }
 
     HIP_CHECK(hipDeviceSynchronize());
     ASSERT_FALSE(out_of_bounds.get());
@@ -324,7 +334,7 @@ TYPED_TEST(RocprimDevicePartitionTests, PredicateEmptyInput)
 
     if (TestFixture::use_graphs)
     {
-        test_utils::cleanupGraphHelper(graph, graph_instance);
+        gHelper.cleanupGraphHelper();
         HIP_CHECK(hipStreamDestroy(stream));
     }
 }
@@ -337,6 +347,7 @@ TYPED_TEST(RocprimDevicePartitionTests, Predicate)
 
     using T = typename TestFixture::input_type;
     using U = typename TestFixture::output_type;
+    using config                                = typename TestFixture::config;
     static constexpr bool use_identity_iterator = TestFixture::use_identity_iterator;
     const bool debug_synchronous = TestFixture::debug_synchronous;
 
@@ -392,15 +403,10 @@ TYPED_TEST(RocprimDevicePartitionTests, Predicate)
             }
             std::reverse(expected_rejected.begin(), expected_rejected.end());
 
-            hipGraph_t graph;
-            hipGraphExec_t graph_instance;
-            if (TestFixture::use_graphs)
-                graph = test_utils::createGraphHelper(stream);
-            
             // temp storage
             size_t temp_storage_size_bytes;
             // Get size of d_temp_storage
-            HIP_CHECK(rocprim::partition(
+            HIP_CHECK(rocprim::partition<config>(
                 nullptr,
                 temp_storage_size_bytes,
                 d_input,
@@ -411,9 +417,6 @@ TYPED_TEST(RocprimDevicePartitionTests, Predicate)
                 stream,
                 debug_synchronous));
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-            
             // temp_storage_size_bytes must be >0
             ASSERT_GT(temp_storage_size_bytes, 0);
 
@@ -421,11 +424,14 @@ TYPED_TEST(RocprimDevicePartitionTests, Predicate)
             void* d_temp_storage = nullptr;
             HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
 
-            if (TestFixture::use_graphs)
-                test_utils::resetGraphHelper(graph, graph_instance, stream);
-            
+            test_utils::GraphHelper gHelper;;
+            if(TestFixture::use_graphs)
+            {
+                gHelper.startStreamCapture(stream);
+            }
+
             // Run
-            HIP_CHECK(rocprim::partition(
+            HIP_CHECK(rocprim::partition<config>(
                 d_temp_storage,
                 temp_storage_size_bytes,
                 d_input,
@@ -436,8 +442,11 @@ TYPED_TEST(RocprimDevicePartitionTests, Predicate)
                 stream,
                 debug_synchronous));
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
+            
+            if(TestFixture::use_graphs)
+            {
+                gHelper.createAndLaunchGraph(stream, true, false);
+            }
 
             HIP_CHECK(hipDeviceSynchronize());
 
@@ -470,13 +479,17 @@ TYPED_TEST(RocprimDevicePartitionTests, Predicate)
             hipFree(d_selected_count_output);
             hipFree(d_temp_storage);
 
-            if (TestFixture::use_graphs)
-                test_utils::cleanupGraphHelper(graph, graph_instance);
+            if(TestFixture::use_graphs)
+            {
+                gHelper.cleanupGraphHelper();
+            }
         }
     }
 
-    if (TestFixture::use_graphs)
+    if(TestFixture::use_graphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
 
 TYPED_TEST(RocprimDevicePartitionTests, PredicateTwoWay)
@@ -487,6 +500,7 @@ TYPED_TEST(RocprimDevicePartitionTests, PredicateTwoWay)
 
     using T                                     = typename TestFixture::input_type;
     using U                                     = typename TestFixture::output_type;
+    using config                                = typename TestFixture::config;
     static constexpr bool use_identity_iterator = TestFixture::use_identity_iterator;
     const bool            debug_synchronous     = TestFixture::debug_synchronous;
 
@@ -548,28 +562,21 @@ TYPED_TEST(RocprimDevicePartitionTests, PredicateTwoWay)
                 }
             }
 
-            hipGraph_t graph;
-            hipGraphExec_t graph_instance;
-            if (TestFixture::use_graphs)
-                graph = test_utils::createGraphHelper(stream);
-            
             // temp storage
             size_t temp_storage_size_bytes;
             // Get size of d_temp_storage
-            HIP_CHECK(rocprim::partition(
+            HIP_CHECK(rocprim::partition_two_way<config>(
                 nullptr,
                 temp_storage_size_bytes,
                 d_input,
                 test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_selected),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_rejected),
                 d_selected_count_output,
                 input.size(),
                 select_op,
                 stream,
                 debug_synchronous));
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-            
             // temp_storage_size_bytes must be >0
             ASSERT_GT(temp_storage_size_bytes, 0);
 
@@ -577,11 +584,14 @@ TYPED_TEST(RocprimDevicePartitionTests, PredicateTwoWay)
             void* d_temp_storage = nullptr;
             HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
 
-            if (TestFixture::use_graphs)
-                test_utils::resetGraphHelper(graph, graph_instance, stream);
-            
+            test_utils::GraphHelper gHelper;;
+            if(TestFixture::use_graphs)
+            {
+                gHelper.startStreamCapture(stream);
+            }
+
             // Run
-            HIP_CHECK(rocprim::partition_two_way(
+            HIP_CHECK(rocprim::partition_two_way<config>(
                 d_temp_storage,
                 temp_storage_size_bytes,
                 d_input,
@@ -593,8 +603,11 @@ TYPED_TEST(RocprimDevicePartitionTests, PredicateTwoWay)
                 stream,
                 debug_synchronous));
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
+            
+            if(TestFixture::use_graphs)
+            {
+                gHelper.createAndLaunchGraph(stream, true, false);
+            }
 
             HIP_CHECK(hipDeviceSynchronize());
 
@@ -629,13 +642,17 @@ TYPED_TEST(RocprimDevicePartitionTests, PredicateTwoWay)
             hipFree(d_selected_count_output);
             hipFree(d_temp_storage);
 
-            if (TestFixture::use_graphs)
-                test_utils::cleanupGraphHelper(graph, graph_instance);
+            if(TestFixture::use_graphs)
+            {
+                gHelper.cleanupGraphHelper();
+            }
         }
     }
 
-    if (TestFixture::use_graphs)
+    if(TestFixture::use_graphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
 
 namespace {
@@ -658,6 +675,7 @@ TYPED_TEST(RocprimDevicePartitionTests, PredicateThreeWay)
 {
     using T = typename TestFixture::input_type;
     using U = typename TestFixture::output_type;
+    using config                                = typename TestFixture::config;
     static constexpr bool use_identity_iterator = TestFixture::use_identity_iterator;
     const bool debug_synchronous = TestFixture::debug_synchronous;
 
@@ -728,21 +746,17 @@ TYPED_TEST(RocprimDevicePartitionTests, PredicateThreeWay)
                     static_cast<unsigned int>(second_partiton_point - partion_point)
                 };
 
-                const auto expected = [&]{
+                const auto expected = [&]
+                {
                     auto result = std::vector<U>(copy.size());
                     std::copy(copy.cbegin(), copy.cend(), result.begin());
                     return result;
                 }();
 
-                hipGraph_t graph;
-                hipGraphExec_t graph_instance;
-                if (TestFixture::use_graphs)
-                    graph = test_utils::createGraphHelper(stream);
-                
                 // temp storage
                 size_t temp_storage_size_bytes;
                 // Get size of d_temp_storage
-                HIP_CHECK(rocprim::partition_three_way(
+                HIP_CHECK(rocprim::partition_three_way<config>(
                     nullptr,
                     temp_storage_size_bytes,
                     d_input,
@@ -757,9 +771,6 @@ TYPED_TEST(RocprimDevicePartitionTests, PredicateThreeWay)
                     stream,
                     debug_synchronous));
 
-                if (TestFixture::use_graphs)
-                    graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-                
                 // temp_storage_size_bytes must be >0
                 ASSERT_GT(temp_storage_size_bytes, 0);
 
@@ -767,11 +778,14 @@ TYPED_TEST(RocprimDevicePartitionTests, PredicateThreeWay)
                 void* d_temp_storage = nullptr;
                 HIP_CHECK(hipMalloc(&d_temp_storage, temp_storage_size_bytes));
 
-                if (TestFixture::use_graphs)
-                    test_utils::resetGraphHelper(graph, graph_instance, stream);
-                
+                test_utils::GraphHelper gHelper;;
+                if(TestFixture::use_graphs)
+                {
+                    gHelper.startStreamCapture(stream);
+                }
+
                 // Run
-                HIP_CHECK(rocprim::partition_three_way(
+                HIP_CHECK(rocprim::partition_three_way<config>(
                     d_temp_storage,
                     temp_storage_size_bytes,
                     d_input,
@@ -786,8 +800,11 @@ TYPED_TEST(RocprimDevicePartitionTests, PredicateThreeWay)
                     stream,
                     debug_synchronous));
 
-                if (TestFixture::use_graphs)
-                    graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
+                
+                if(TestFixture::use_graphs)
+                {
+                    gHelper.createAndLaunchGraph(stream);
+                }
 
                 HIP_CHECK(hipDeviceSynchronize());
 
@@ -838,14 +855,18 @@ TYPED_TEST(RocprimDevicePartitionTests, PredicateThreeWay)
                 hipFree(d_selected_counts);
                 hipFree(d_temp_storage);
 
-                if (TestFixture::use_graphs)
-                    test_utils::cleanupGraphHelper(graph, graph_instance);
+                if(TestFixture::use_graphs)
+                {
+                    gHelper.cleanupGraphHelper();
+                }
             }
         }
     }
 
-    if (TestFixture::use_graphs)
+    if(TestFixture::use_graphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
 
 namespace
@@ -1130,11 +1151,6 @@ TEST_P(RocprimDevicePartitionLargeInputTests, LargeInputPartition)
         size_t* d_count_output{};
         HIP_CHECK(test_common_utils::hipMallocHelper(&d_count_output, sizeof(*d_count_output)));
 
-        hipGraph_t graph;
-        hipGraphExec_t graph_instance;
-        if (use_graphs)
-            graph = test_utils::createGraphHelper(stream);
-        
         void*  d_temporary_storage{};
         size_t temporary_storage_size{};
         HIP_CHECK(rocprim::partition(d_temporary_storage,
@@ -1147,15 +1163,15 @@ TEST_P(RocprimDevicePartitionLargeInputTests, LargeInputPartition)
                                      stream,
                                      debug_synchronous));
 
-        if (use_graphs)
-            graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-        
         ASSERT_NE(0, temporary_storage_size);
         HIP_CHECK(test_common_utils::hipMallocHelper(&d_temporary_storage, temporary_storage_size));
 
-        if (use_graphs)
-            test_utils::resetGraphHelper(graph, graph_instance, stream);
-        
+        test_utils::GraphHelper gHelper;;
+        if(use_graphs)
+        {
+            gHelper.startStreamCapture(stream);
+        }
+
         HIP_CHECK(rocprim::partition(d_temporary_storage,
                                      temporary_storage_size,
                                      input_iterator,
@@ -1166,9 +1182,12 @@ TEST_P(RocprimDevicePartitionLargeInputTests, LargeInputPartition)
                                      stream,
                                      debug_synchronous));
 
-        if (use_graphs)
-            graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
         
+        if(use_graphs)
+        {
+            gHelper.createAndLaunchGraph(stream);
+        }
+
         size_t count_output{};
         HIP_CHECK(hipMemcpyWithStream(&count_output,
                                       d_count_output,
@@ -1192,12 +1211,16 @@ TEST_P(RocprimDevicePartitionLargeInputTests, LargeInputPartition)
         HIP_CHECK(hipFree(d_count_output));
         HIP_CHECK(hipFree(d_incorrect_flag));
 
-        if (use_graphs)
-            test_utils::cleanupGraphHelper(graph, graph_instance);
+        if(use_graphs)
+        {
+            gHelper.cleanupGraphHelper();
+        }
     }
 
-    if (use_graphs)
+    if(use_graphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
 
 TEST_P(RocprimDevicePartitionLargeInputTests, LargeInputPartitionTwoWay)
@@ -1248,11 +1271,6 @@ TEST_P(RocprimDevicePartitionLargeInputTests, LargeInputPartitionTwoWay)
         size_t* d_count_output{};
         HIP_CHECK(test_common_utils::hipMallocHelper(&d_count_output, sizeof(*d_count_output)));
 
-        hipGraph_t graph;
-        hipGraphExec_t graph_instance;
-        if (use_graphs)
-            graph = test_utils::createGraphHelper(stream);
-        
         void*  d_temporary_storage{};
         size_t temporary_storage_size{};
         HIP_CHECK(rocprim::partition_two_way(d_temporary_storage,
@@ -1266,15 +1284,15 @@ TEST_P(RocprimDevicePartitionLargeInputTests, LargeInputPartitionTwoWay)
                                              stream,
                                              debug_synchronous));
 
-        if (use_graphs)
-            graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-        
         ASSERT_NE(0, temporary_storage_size);
         HIP_CHECK(test_common_utils::hipMallocHelper(&d_temporary_storage, temporary_storage_size));
 
-        if (use_graphs)
-            test_utils::resetGraphHelper(graph, graph_instance, stream);
-        
+        test_utils::GraphHelper gHelper;;
+        if(use_graphs)
+        {
+            gHelper.startStreamCapture(stream);
+        }
+
         HIP_CHECK(rocprim::partition_two_way(d_temporary_storage,
                                              temporary_storage_size,
                                              input_iterator,
@@ -1286,9 +1304,12 @@ TEST_P(RocprimDevicePartitionLargeInputTests, LargeInputPartitionTwoWay)
                                              stream,
                                              debug_synchronous));
 
-        if (use_graphs)
-            graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
         
+        if(use_graphs)
+        {
+            gHelper.createAndLaunchGraph(stream);
+        }
+
         size_t count_output{};
         HIP_CHECK(hipMemcpyWithStream(&count_output,
                                       d_count_output,
@@ -1320,12 +1341,16 @@ TEST_P(RocprimDevicePartitionLargeInputTests, LargeInputPartitionTwoWay)
         HIP_CHECK(hipFree(d_incorrect_select_flag));
         HIP_CHECK(hipFree(d_incorrect_reject_flag));
 
-        if (use_graphs)
-            test_utils::cleanupGraphHelper(graph, graph_instance);
+        if(use_graphs)
+        {
+            gHelper.cleanupGraphHelper();
+        }
     }
 
-    if (use_graphs)
+    if(use_graphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
 
 TEST_P(RocprimDevicePartitionLargeInputTests, LargeInputPartitionThreeWay)
@@ -1373,11 +1398,6 @@ TEST_P(RocprimDevicePartitionLargeInputTests, LargeInputPartitionThreeWay)
         size_t* d_count_output{};
         HIP_CHECK(test_common_utils::hipMallocHelper(&d_count_output, 2 * sizeof(*d_count_output)));
 
-        hipGraph_t graph;
-        hipGraphExec_t graph_instance;
-        if (use_graphs)
-            graph = test_utils::createGraphHelper(stream);
-        
         void*  d_temporary_storage{};
         size_t temporary_storage_size{};
         HIP_CHECK(rocprim::partition_three_way(d_temporary_storage,
@@ -1393,15 +1413,15 @@ TEST_P(RocprimDevicePartitionLargeInputTests, LargeInputPartitionThreeWay)
                                                stream,
                                                debug_synchronous));
 
-        if (use_graphs)
-            graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-        
         ASSERT_NE(0, temporary_storage_size);
         HIP_CHECK(test_common_utils::hipMallocHelper(&d_temporary_storage, temporary_storage_size));
 
-        if (use_graphs)
-            test_utils::resetGraphHelper(graph, graph_instance, stream);
-        
+        test_utils::GraphHelper gHelper;;
+        if(use_graphs)
+        {
+            gHelper.startStreamCapture(stream);
+        }
+
         HIP_CHECK(rocprim::partition_three_way(d_temporary_storage,
                                                temporary_storage_size,
                                                input_iterator,
@@ -1415,9 +1435,12 @@ TEST_P(RocprimDevicePartitionLargeInputTests, LargeInputPartitionThreeWay)
                                                stream,
                                                debug_synchronous));
 
-        if (use_graphs)
-            graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
         
+        if(use_graphs)
+        {
+            gHelper.createAndLaunchGraph(stream);
+        }
+
         size_t count_output[2]{};
         HIP_CHECK(hipMemcpyWithStream(&count_output,
                                       d_count_output,
@@ -1445,10 +1468,171 @@ TEST_P(RocprimDevicePartitionLargeInputTests, LargeInputPartitionThreeWay)
         HIP_CHECK(hipFree(d_count_output));
         HIP_CHECK(hipFree(d_incorrect_flag));
 
-        if (use_graphs)
-            test_utils::cleanupGraphHelper(graph, graph_instance);
+        if(use_graphs)
+        {
+            gHelper.cleanupGraphHelper();
+        }
     }
 
-    if (use_graphs)
+    if(use_graphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
+
+// This test checks to make sure that the block size is reduced correctly
+// when our data size and type are set in a way that we will exceed the shared
+// memory limit. Since the block size calculation is done at compile time,
+// if the block size is not correctly reduced, this test will fail to compile.
+TEST(RocprimDevicePartitionBlockSizeTests, BlockSize)
+{
+    int device_id = test_common_utils::obtain_device_from_ctest();
+    SCOPED_TRACE(testing::Message() << "with device_id = " << device_id);
+    HIP_CHECK(hipSetDevice(device_id));
+
+    // Create a large struct to test with. It must be big enough that when
+    // we the use default block size (defined in rocprim::default_select_config 
+    // struct as 256), giving one instance to each thread will cause us to hit
+    // 32 KiB of shared memory (the limit enforced by the rocprim::limit_block_size
+    // struct's boolean template parameter). Since the device_partition algorithm also
+    // uses some shared memory to store state, this will cause the total usage to exceed
+    // the 32 KiB limit. If everything's working correctly, this should be detected in 
+    // the limit_block_size's template logic, and it should reduce the block size.
+    const size_t test_obj_size = 128; // Choose 128, since 256 * 128 = 2^15 bytes (32 KiB).
+    struct TestObject
+    {
+        unsigned char data[test_obj_size];
+
+        bool operator==(const TestObject& other) const
+        {
+            bool equal = true;
+            for (size_t i = 0; equal && i < test_obj_size; i++)
+                equal = data[i] == other.data[i];
+
+            return equal;
+        }
+    };
+
+    using T = TestObject; // input data type
+    using U = TestObject; // output data type
+    const bool debug_synchronous = false;
+    const hipStream_t stream = 0; // default stream
+
+    auto select_op = [] __host__ __device__ (const T& value) -> bool
+    {
+        // The data values are in [0, 255]. Partition on the midpoint.
+        if(value.data[0] == 128) return true;
+        return false;
+    };
+
+    // Use some power of two and off-by-one-from-power-of-two data sizes.
+    const std::vector<size_t> sizes = {256, 257, 511, 512, 1024, 1025};
+
+    for (size_t seed_index = 0; seed_index < random_seeds_count + seed_size; seed_index++)
+    {
+        unsigned int seed_value = seed_index < random_seeds_count  ? rand() : seeds[seed_index - random_seeds_count];
+        SCOPED_TRACE(testing::Message() << "with seed = " << seed_value);
+
+        for(auto size : sizes)
+        {
+            SCOPED_TRACE(testing::Message() << "with size = " << size);
+
+            // Generate data
+            std::vector<unsigned char> input_data = test_utils::get_random_data<unsigned char>(size * test_obj_size, 0, 255, seed_value);
+            std::vector<T> input(size);
+            for (size_t i = 0; i < size; i++)
+                memcpy(input[i].data, input_data.data() + i * test_obj_size, test_obj_size);
+
+            T * d_input;
+            U * d_output;
+            unsigned int * d_selected_count_output;
+            HIP_CHECK(test_common_utils::hipMallocHelper(&d_input, input.size() * sizeof(T)));
+            HIP_CHECK(test_common_utils::hipMallocHelper(&d_output, input.size() * sizeof(U)));
+            HIP_CHECK(test_common_utils::hipMallocHelper(&d_selected_count_output, sizeof(unsigned int)));
+            HIP_CHECK(
+                hipMemcpy(d_input, input.data(), input.size() * sizeof(T), hipMemcpyHostToDevice));
+
+            // Calculate expected_selected and expected_rejected results on host
+            std::vector<U> expected_selected;
+            std::vector<U> expected_rejected;
+            expected_selected.reserve(input.size()/2);
+            expected_rejected.reserve(input.size()/2);
+            for(size_t i = 0; i < input.size(); i++)
+            {
+                if(select_op(input[i]))
+                {
+                    expected_selected.push_back(input[i]);
+                }
+                else
+                {
+                    expected_rejected.push_back(input[i]);
+                }
+            }
+            std::reverse(expected_rejected.begin(), expected_rejected.end());
+
+            // temp storage
+            size_t temp_storage_size_bytes;
+            // Get size of d_temp_storage
+            HIP_CHECK(rocprim::partition(
+                nullptr,
+                temp_storage_size_bytes,
+                d_input,
+                d_output,
+                d_selected_count_output,
+                input.size(),
+                select_op,
+                stream,
+                debug_synchronous));
+
+            // temp_storage_size_bytes must be >0
+            ASSERT_GT(temp_storage_size_bytes, 0);
+
+            // allocate temporary storage
+            void* d_temp_storage = nullptr;
+            HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
+
+            // Run
+            HIP_CHECK(rocprim::partition(
+                d_temp_storage,
+                temp_storage_size_bytes,
+                d_input,
+                d_output,
+                d_selected_count_output,
+                input.size(),
+                select_op,
+                stream,
+                debug_synchronous));
+
+            HIP_CHECK(hipDeviceSynchronize());
+
+            // Check if number of selected value is as expected_selected
+            unsigned int selected_count_output = 0;
+            HIP_CHECK(hipMemcpy(&selected_count_output,
+                                d_selected_count_output,
+                                sizeof(unsigned int),
+                                hipMemcpyDeviceToHost));
+            ASSERT_EQ(selected_count_output, expected_selected.size());
+
+            // Check if output values are as expected_selected
+            std::vector<U> output(input.size());
+            HIP_CHECK(hipMemcpy(output.data(),
+                                d_output,
+                                output.size() * sizeof(U),
+                                hipMemcpyDeviceToHost));
+
+            std::vector<U> output_rejected;
+            for(size_t i = 0; i < expected_rejected.size(); i++)
+            {
+                auto j = i + expected_selected.size();
+                output_rejected.push_back(output[j]);
+            }
+            ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(output, expected_selected, expected_selected.size()));
+            ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(output_rejected, expected_rejected, expected_rejected.size()));
+
+            hipFree(d_input);
+            hipFree(d_output);
+            hipFree(d_selected_count_output);
+            hipFree(d_temp_storage);
+        }
+    }
+}
\ No newline at end of file
