@@ -1,6 +1,6 @@
 // MIT License
 //
-// Copyright (c) 2017-2023 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -39,7 +39,8 @@ template<class InputType,
          bool   UseIdentityIterator = false,
          size_t SizeLimit           = ROCPRIM_GRID_SIZE_LIMIT,
          bra    Algo                = bra::default_algorithm,
-         bool   UseGraphs           = false>
+         bool   UseGraphs           = false,
+         bool   Deterministic       = false>
 struct DeviceReduceParams
 {
     static constexpr bra algo = Algo;
@@ -98,25 +99,51 @@ typedef ::testing::Types<
     DeviceReduceParams<int, float>,
     DeviceReduceParamsList(int, int, false, 512),
     DeviceReduceParamsList(float, float, false, 2048),
+    DeviceReduceParamsList(double, double, false, 2048),
     DeviceReduceParamsList(int, int, false, 4096),
     DeviceReduceParamsList(int, int, false, 2097152),
     DeviceReduceParamsList(int, int, false, 1073741824),
     DeviceReduceParams<int8_t, int8_t>,
     DeviceReduceParams<uint8_t, uint8_t>,
-    // #156 temporarily disable half test due to known issue with converting from double to half
-    // DeviceReduceParams<rocprim::half, rocprim::half>,
+    DeviceReduceParams<rocprim::half, rocprim::half>,
     DeviceReduceParams<rocprim::bfloat16, rocprim::bfloat16>,
     DeviceReduceParams<test_utils::custom_test_type<float>, test_utils::custom_test_type<float>>,
     DeviceReduceParams<test_utils::custom_test_type<int>, test_utils::custom_test_type<float>>,
+    DeviceReduceParams<rocprim::half,
+                       rocprim::half,
+                       false,
+                       ROCPRIM_GRID_SIZE_LIMIT,
+                       bra::default_algorithm,
+                       false,
+                       true>,
+    DeviceReduceParams<float,
+                       float,
+                       false,
+                       ROCPRIM_GRID_SIZE_LIMIT,
+                       bra::default_algorithm,
+                       false,
+                       true>,
+    DeviceReduceParams<double,
+                       double,
+                       false,
+                       ROCPRIM_GRID_SIZE_LIMIT,
+                       bra::default_algorithm,
+                       false,
+                       true>,
+    DeviceReduceParams<test_utils::custom_test_type<double>,
+                       test_utils::custom_test_type<double>,
+                       false,
+                       ROCPRIM_GRID_SIZE_LIMIT,
+                       bra::default_algorithm,
+                       false,
+                       true>,
     DeviceReduceParams<int, int, false, ROCPRIM_GRID_SIZE_LIMIT, bra::default_algorithm, true>>
     RocprimDeviceReduceTestsParams;
 
-typedef ::testing::Types<
-    DeviceReduceParams<double, double>,
-    DeviceReduceParamsList(float, float, false, 2048),
-    // #156 temporarily disable half test due to known issue with converting from double to half
-    // DeviceReduceParams<rocprim::half, rocprim::half>,
-    DeviceReduceParams<rocprim::bfloat16, rocprim::bfloat16>>
+typedef ::testing::Types<DeviceReduceParams<double, double>,
+                         DeviceReduceParamsList(float, float, false, 2048),
+                         DeviceReduceParams<rocprim::half, rocprim::half>,
+                         DeviceReduceParams<rocprim::bfloat16, rocprim::bfloat16>>
     RocprimDeviceReducePrecisionTestsParams;
 
 TYPED_TEST_SUITE(RocprimDeviceReduceTests, RocprimDeviceReduceTestsParams);
@@ -145,11 +172,6 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceEmptyInput)
 
     const U initial_value = U(1234);
 
-    hipGraph_t graph;
-    hipGraphExec_t graph_instance;
-    if (TestFixture::use_graphs)
-        graph = test_utils::createGraphHelper(stream);
-    
     size_t temp_storage_size_bytes;
     // Get size of d_temp_storage
     HIP_CHECK(
@@ -162,14 +184,14 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceEmptyInput)
         )
     );
 
-    if (TestFixture::use_graphs)
-        graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-
     void * d_temp_storage = nullptr;
     HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
 
-    if (TestFixture::use_graphs)
-        test_utils::resetGraphHelper(graph, graph_instance, stream);
+    test_utils::GraphHelper gHelper;
+    if(TestFixture::use_graphs)
+    {
+        gHelper.startStreamCapture(stream);
+    }
 
     // Run
     HIP_CHECK(
@@ -182,9 +204,12 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceEmptyInput)
         )
     );
 
-    if (TestFixture::use_graphs)
-        graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
     
+    if(TestFixture::use_graphs)
+    {
+        gHelper.createAndLaunchGraph(stream);
+    }
+
     HIP_CHECK(hipDeviceSynchronize());
 
     U output;
@@ -199,10 +224,10 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceEmptyInput)
 
     hipFree(d_output);
     hipFree(d_temp_storage);
-    
+
     if (TestFixture::use_graphs)
     {
-        test_utils::cleanupGraphHelper(graph, graph_instance);
+        gHelper.cleanupGraphHelper();
         HIP_CHECK(hipStreamDestroy(stream));
     }
 }
@@ -269,27 +294,20 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceSum)
             if(size == 0)
                 expected = U();
 
-            hipGraph_t graph;
-            hipGraphExec_t graph_instance;
-            if (TestFixture::use_graphs)
-                graph = test_utils::createGraphHelper(stream);
-            
             // temp storage
             size_t temp_storage_size_bytes;
             void * d_temp_storage = nullptr;
             // Get size of d_temp_storage
-            HIP_CHECK(
-                rocprim::reduce<Config>(
-                    d_temp_storage, temp_storage_size_bytes,
-                    d_input,
-                    test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output),
-                    input.size(), rocprim::plus<U>(), stream, debug_synchronous
-                )
-            );
+            HIP_CHECK(rocprim::reduce<Config>(
+                d_temp_storage,
+                temp_storage_size_bytes,
+                d_input,
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output),
+                input.size(),
+                rocprim::plus<U>(),
+                stream,
+                debug_synchronous));
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-            
             // temp_storage_size_bytes must be >0
             ASSERT_GT(temp_storage_size_bytes, 0);
 
@@ -297,9 +315,12 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceSum)
             HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
             HIP_CHECK(hipDeviceSynchronize());
 
-            if (TestFixture::use_graphs)
-                test_utils::resetGraphHelper(graph, graph_instance, stream);
-            
+            test_utils::GraphHelper gHelper;
+            if(TestFixture::use_graphs)
+            {
+                gHelper.startStreamCapture(stream);
+            }
+
             // Run
             HIP_CHECK(
                 rocprim::reduce<Config>(
@@ -310,8 +331,11 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceSum)
                 )
             );
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
+            
+            if(TestFixture::use_graphs)
+            {
+                gHelper.createAndLaunchGraph(stream);
+            }
 
             HIP_CHECK(hipGetLastError());
             HIP_CHECK(hipDeviceSynchronize());
@@ -333,10 +357,10 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceSum)
             hipFree(d_input);
             hipFree(d_output);
             hipFree(d_temp_storage);
-            
+
             if (TestFixture::use_graphs)
             {
-                test_utils::cleanupGraphHelper(graph, graph_instance);
+                gHelper.cleanupGraphHelper();
                 HIP_CHECK(hipStreamDestroy(stream));
             }
         }
@@ -423,11 +447,6 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceArgMinimum)
                 expected = reduce_op(expected, input[i]);
             }
 
-            hipGraph_t graph;
-            hipGraphExec_t graph_instance;
-            if (TestFixture::use_graphs)
-                graph = test_utils::createGraphHelper(stream);
-
             // temp storage
             size_t temp_storage_size_bytes;
             void * d_temp_storage = nullptr;
@@ -441,9 +460,6 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceArgMinimum)
                 )
             );
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-
             // temp_storage_size_bytes must be >0
             ASSERT_GT(temp_storage_size_bytes, 0);
 
@@ -451,9 +467,12 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceArgMinimum)
             HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
             HIP_CHECK(hipDeviceSynchronize());
 
-            if (TestFixture::use_graphs)
-                test_utils::resetGraphHelper(graph, graph_instance, stream);
-            
+            test_utils::GraphHelper gHelper;
+            if(TestFixture::use_graphs)
+            {
+                gHelper.startStreamCapture(stream);
+            }
+
             // Run
             HIP_CHECK(
                 rocprim::reduce<Config>(
@@ -464,9 +483,12 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceArgMinimum)
                 )
             );
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
             
+            if(TestFixture::use_graphs)
+            {
+                gHelper.createAndLaunchGraph(stream);
+            }
+
             HIP_CHECK(hipGetLastError());
             HIP_CHECK(hipDeviceSynchronize());
 
@@ -490,7 +512,7 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceArgMinimum)
 
             if (TestFixture::use_graphs)
             {
-                test_utils::cleanupGraphHelper(graph, graph_instance);
+                gHelper.cleanupGraphHelper();
                 HIP_CHECK(hipStreamDestroy(stream));
             }
         }
@@ -530,11 +552,6 @@ void testLargeIndices()
             T* d_output = nullptr;
             HIP_CHECK(test_common_utils::hipMallocHelper(&d_output, sizeof(T)));
 
-            hipGraph_t graph;
-            hipGraphExec_t graph_instance;
-            if (use_graphs)
-                graph = test_utils::createGraphHelper(stream);
-            
             // temp storage
             size_t temp_storage_size_bytes = 0;
             void*  d_temp_storage          = nullptr;
@@ -544,20 +561,20 @@ void testLargeIndices()
                                       input,
                                       d_output,
                                       size,
-                                      rocprim::plus<T> {},
+                                      rocprim::plus<T>{},
                                       stream,
                                       debug_synchronous));
 
-            if (use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-            
             // allocate temporary storage
             HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
             HIP_CHECK(hipDeviceSynchronize());
 
-            if (use_graphs)
-                test_utils::resetGraphHelper(graph, graph_instance, stream);
-            
+            test_utils::GraphHelper gHelper;
+            if(use_graphs)
+            {
+                gHelper.startStreamCapture(stream);
+            }
+
             // Run
             HIP_CHECK(rocprim::reduce(d_temp_storage,
                                       temp_storage_size_bytes,
@@ -568,9 +585,12 @@ void testLargeIndices()
                                       stream,
                                       debug_synchronous));
 
-            if (use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
             
+            if(use_graphs)
+            {
+                gHelper.createAndLaunchGraph(stream, true, false);
+            }
+
             HIP_CHECK(hipGetLastError());
             HIP_CHECK(hipDeviceSynchronize());
 
@@ -588,13 +608,17 @@ void testLargeIndices()
             hipFree(d_temp_storage);
             hipFree(d_output);
 
-            if (use_graphs)
-                test_utils::cleanupGraphHelper(graph, graph_instance);
+            if(use_graphs)
+            {
+                gHelper.cleanupGraphHelper();
+            }
         }
     }
 
-    if (use_graphs)
+    if(use_graphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
 
 TEST(RocprimDeviceReduceTests, LargeIndices)
@@ -664,13 +688,8 @@ TYPED_TEST(RocprimDeviceReducePrecisionTests, ReduceSumInputEqualExponentFunctio
         HIP_CHECK(hipDeviceSynchronize());
 
         // Calculate expected results on host mathematically (instead of using reduce on host)
-        U expected  = static_cast<U>(static_cast<double>(size) * static_cast<double>(lowest));
+        U expected = static_cast<U>(static_cast<double>(size) * static_cast<double>(lowest));
 
-        hipGraph_t graph;
-        hipGraphExec_t graph_instance;
-        if (TestFixture::use_graphs)
-            graph = test_utils::createGraphHelper(stream);
-        
         // temp storage
         size_t temp_storage_size_bytes;
         void*  d_temp_storage = nullptr;
@@ -685,9 +704,6 @@ TYPED_TEST(RocprimDeviceReducePrecisionTests, ReduceSumInputEqualExponentFunctio
             stream,
             debug_synchronous));
 
-        if (TestFixture::use_graphs)
-            graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-        
         // temp_storage_size_bytes must be >0
         ASSERT_GT(temp_storage_size_bytes, 0);
 
@@ -695,9 +711,12 @@ TYPED_TEST(RocprimDeviceReducePrecisionTests, ReduceSumInputEqualExponentFunctio
         HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
         HIP_CHECK(hipDeviceSynchronize());
 
-        if (TestFixture::use_graphs)
-            test_utils::resetGraphHelper(graph, graph_instance, stream);
-        
+        test_utils::GraphHelper gHelper;
+        if(TestFixture::use_graphs)
+        {
+            gHelper.startStreamCapture(stream);
+        }
+
         // Run
         HIP_CHECK(rocprim::reduce<Config>(
             d_temp_storage,
@@ -709,9 +728,12 @@ TYPED_TEST(RocprimDeviceReducePrecisionTests, ReduceSumInputEqualExponentFunctio
             stream,
             debug_synchronous));
 
-        if (TestFixture::use_graphs)
-            graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
         
+        if(TestFixture::use_graphs)
+        {
+            gHelper.createAndLaunchGraph(stream);
+        }
+
         HIP_CHECK(hipGetLastError());
         HIP_CHECK(hipDeviceSynchronize());
 
@@ -729,7 +751,7 @@ TYPED_TEST(RocprimDeviceReducePrecisionTests, ReduceSumInputEqualExponentFunctio
 
         if (TestFixture::use_graphs)
         {
-            test_utils::cleanupGraphHelper(graph, graph_instance);
+            gHelper.cleanupGraphHelper();
             HIP_CHECK(hipStreamDestroy(stream));
         }
     }
@@ -792,11 +814,6 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceMinimum)
                 expected = min_op(expected, input[i]);
             }
 
-            hipGraph_t graph;
-            hipGraphExec_t graph_instance;
-            if (TestFixture::use_graphs)
-                graph = test_utils::createGraphHelper(stream);
-
             // Get size of d_temp_storage
             size_t temp_storage_size_bytes;
             void * d_temp_storage = nullptr;
@@ -809,9 +826,6 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceMinimum)
                 )
             );
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-
             // temp_storage_size_bytes must be >0
             ASSERT_GT(temp_storage_size_bytes, 0);
 
@@ -819,8 +833,11 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceMinimum)
             HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
             HIP_CHECK(hipDeviceSynchronize());
 
-            if (TestFixture::use_graphs)
-                test_utils::resetGraphHelper(graph, graph_instance, stream);
+            test_utils::GraphHelper gHelper;
+            if(TestFixture::use_graphs)
+            {
+                gHelper.startStreamCapture(stream);
+            }
 
             // Run
             HIP_CHECK(
@@ -841,9 +858,12 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceMinimum)
                 )
             );
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
             
+            if(TestFixture::use_graphs)
+            {
+                gHelper.createAndLaunchGraph(stream, true, false);
+            }
+
             HIP_CHECK(hipDeviceSynchronize());
 
             // Check if output values are as expected
@@ -860,7 +880,7 @@ TYPED_TEST(RocprimDeviceReduceTests, ReduceMinimum)
 
             if (TestFixture::use_graphs)
             {
-                test_utils::cleanupGraphHelper(graph, graph_instance);
+                gHelper.cleanupGraphHelper();
                 HIP_CHECK(hipStreamDestroy(stream));
             }
         }
