@@ -1,6 +1,6 @@
 // MIT License
 //
-// Copyright (c) 2017-2022 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -61,17 +61,20 @@ public:
     static constexpr bool use_graphs = Params::use_graphs;
 };
 
-typedef ::testing::Types<
-    DeviceSelectParams<int, long>,
-    DeviceSelectParams<int8_t, int8_t>,
-    DeviceSelectParams<uint8_t, uint8_t>,
-    DeviceSelectParams<rocprim::half, rocprim::half>,
-    DeviceSelectParams<rocprim::bfloat16, rocprim::bfloat16>,
-    DeviceSelectParams<unsigned char, float, int, true>,
-    DeviceSelectParams<double, double, int, true>,
-    DeviceSelectParams<test_utils::custom_test_type<double>, test_utils::custom_test_type<double>, int, true>,
-    DeviceSelectParams<int, int, unsigned int, false, true>
-> RocprimDeviceSelectTestsParams;
+typedef ::testing::Types<DeviceSelectParams<int, long>,
+                         DeviceSelectParams<int8_t, int8_t>,
+                         DeviceSelectParams<uint8_t, uint8_t>,
+                         DeviceSelectParams<rocprim::half, rocprim::half>,
+                         DeviceSelectParams<rocprim::bfloat16, rocprim::bfloat16>,
+                         DeviceSelectParams<float, float>,
+                         DeviceSelectParams<unsigned char, float, int, true>,
+                         DeviceSelectParams<double, double, int, true>,
+                         DeviceSelectParams<test_utils::custom_test_type<double>,
+                                            test_utils::custom_test_type<double>,
+                                            int,
+                                            true>,
+                         DeviceSelectParams<int, int, unsigned int, false, true>>
+    RocprimDeviceSelectTestsParams;
 
 TYPED_TEST_SUITE(RocprimDeviceSelectTests, RocprimDeviceSelectTestsParams);
 
@@ -141,31 +144,20 @@ TYPED_TEST(RocprimDeviceSelectTests, Flagged)
                 }
             }
 
-            hipGraph_t graph;
-            hipGraphExec_t graph_instance;
-            if (TestFixture::use_graphs)
-                graph = test_utils::createGraphHelper(stream);
-            
             // temp storage
             size_t temp_storage_size_bytes;
             // Get size of d_temp_storage
-            HIP_CHECK(
-                rocprim::select(
-                    nullptr,
-                    temp_storage_size_bytes,
-                    d_input,
-                    d_flags,
-                    test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output),
-                    d_selected_count_output,
-                    input.size(),
-                    stream,
-                    TestFixture::debug_synchronous
-                )
-            );
+            HIP_CHECK(rocprim::select(
+                nullptr,
+                temp_storage_size_bytes,
+                d_input,
+                d_flags,
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output),
+                d_selected_count_output,
+                input.size(),
+                stream,
+                TestFixture::debug_synchronous));
 
-            if (TestFixture::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
-            
             HIP_CHECK(hipDeviceSynchronize());
 
             // temp_storage_size_bytes must be >0
@@ -176,9 +168,12 @@ TYPED_TEST(RocprimDeviceSelectTests, Flagged)
             HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
             HIP_CHECK(hipDeviceSynchronize());
 
-            if (TestFixture::use_graphs)
-                test_utils::resetGraphHelper(graph, graph_instance, stream);
-            
+            test_utils::GraphHelper gHelper;;
+            if(TestFixture::use_graphs)
+            {
+                gHelper.startStreamCapture(stream);
+            }
+
             // Run
             HIP_CHECK(
                 rocprim::select(
@@ -194,9 +189,12 @@ TYPED_TEST(RocprimDeviceSelectTests, Flagged)
                 )
             );
 
-            if (TestFixture::use_graphs)
-                graph_instance = graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
             
+            if(TestFixture::use_graphs)
+            {
+                gHelper.createAndLaunchGraph(stream, true, false);
+            }
+
             HIP_CHECK(hipDeviceSynchronize());
 
             // Check if number of selected value is as expected
@@ -229,13 +227,17 @@ TYPED_TEST(RocprimDeviceSelectTests, Flagged)
             hipFree(d_selected_count_output);
             hipFree(d_temp_storage);
 
-            if (TestFixture::use_graphs)
-                test_utils::cleanupGraphHelper(graph, graph_instance);
+            if(TestFixture::use_graphs)
+            {
+                gHelper.cleanupGraphHelper();
+            }
         }
     }
 
-    if (TestFixture::use_graphs)
+    if(TestFixture::use_graphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
 
 
@@ -305,31 +307,20 @@ TYPED_TEST(RocprimDeviceSelectTests, SelectOp)
                 }
             }
 
-            hipGraph_t graph;
-            hipGraphExec_t graph_instance;
-            if (TestFixture::use_graphs)
-                graph = test_utils::createGraphHelper(stream);
-
             // temp storage
             size_t temp_storage_size_bytes;
             // Get size of d_temp_storage
-            HIP_CHECK(
-                rocprim::select(
-                    nullptr,
-                    temp_storage_size_bytes,
-                    d_input,
-                    test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output),
-                    d_selected_count_output,
-                    input.size(),
-                    select_op<T>(),
-                    stream,
-                    debug_synchronous
-                )
-            );
+            HIP_CHECK(rocprim::select(
+                nullptr,
+                temp_storage_size_bytes,
+                d_input,
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output),
+                d_selected_count_output,
+                input.size(),
+                select_op<T>(),
+                stream,
+                debug_synchronous));
 
-            if (TestFixture::use_graphs)
-                graph_instance = graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
-            
             HIP_CHECK(hipDeviceSynchronize());
 
             // temp_storage_size_bytes must be >0
@@ -340,8 +331,11 @@ TYPED_TEST(RocprimDeviceSelectTests, SelectOp)
             HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
             HIP_CHECK(hipDeviceSynchronize());
 
-            if (TestFixture::use_graphs)
-                test_utils::resetGraphHelper(graph, graph_instance, stream);
+            test_utils::GraphHelper gHelper;;
+            if(TestFixture::use_graphs)
+            {
+                gHelper.startStreamCapture(stream);
+            }
 
             // Run
             HIP_CHECK(
@@ -358,8 +352,11 @@ TYPED_TEST(RocprimDeviceSelectTests, SelectOp)
                 )
             );
 
-            if (TestFixture::use_graphs)
-                graph_instance = graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
+            
+            if(TestFixture::use_graphs)
+            {
+                gHelper.createAndLaunchGraph(stream, true, false);
+            }
 
             HIP_CHECK(hipDeviceSynchronize());
 
@@ -392,13 +389,17 @@ TYPED_TEST(RocprimDeviceSelectTests, SelectOp)
             hipFree(d_selected_count_output);
             hipFree(d_temp_storage);
 
-            if (TestFixture::use_graphs)
-                test_utils::cleanupGraphHelper(graph, graph_instance);
+            if(TestFixture::use_graphs)
+            {
+                gHelper.cleanupGraphHelper();
+            }
         }
     }
 
-    if (TestFixture::use_graphs)
+    if(TestFixture::use_graphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
 
 std::vector<float> get_discontinuity_probabilities()
@@ -484,31 +485,20 @@ TYPED_TEST(RocprimDeviceSelectTests, Unique)
                     }
                 }
 
-                hipGraph_t graph;
-                hipGraphExec_t graph_instance;
-                if (TestFixture::use_graphs)
-                    graph = test_utils::createGraphHelper(stream);
-                
                 // temp storage
                 size_t temp_storage_size_bytes;
                 // Get size of d_temp_storage
-                HIP_CHECK(
-                    rocprim::unique(
-                        nullptr,
-                        temp_storage_size_bytes,
-                        d_input,
-                        test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output),
-                        d_selected_count_output,
-                        input.size(),
-                        op_type(),
-                        stream,
-                        debug_synchronous
-                    )
-                );
+                HIP_CHECK(rocprim::unique(
+                    nullptr,
+                    temp_storage_size_bytes,
+                    d_input,
+                    test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_output),
+                    d_selected_count_output,
+                    input.size(),
+                    op_type(),
+                    stream,
+                    debug_synchronous));
 
-                if (TestFixture::use_graphs)
-                    graph_instance = graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
-                
                 HIP_CHECK(hipDeviceSynchronize());
 
                 // temp_storage_size_bytes must be >0
@@ -519,8 +509,11 @@ TYPED_TEST(RocprimDeviceSelectTests, Unique)
                 HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
                 HIP_CHECK(hipDeviceSynchronize());
 
-                if (TestFixture::use_graphs)
-                    test_utils::resetGraphHelper(graph, graph_instance, stream);
+                test_utils::GraphHelper gHelper;;
+                if(TestFixture::use_graphs)
+                {
+                    gHelper.startStreamCapture(stream);
+                }
 
                 // Run
                 HIP_CHECK(
@@ -537,9 +530,12 @@ TYPED_TEST(RocprimDeviceSelectTests, Unique)
                     )
                 );
 
-                if (TestFixture::use_graphs)
-                    graph_instance = graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
                 
+                if(TestFixture::use_graphs)
+                {
+                    gHelper.createAndLaunchGraph(stream, true, false);
+                }
+
                 HIP_CHECK(hipDeviceSynchronize());
 
                 // Check if number of selected value is as expected
@@ -571,14 +567,18 @@ TYPED_TEST(RocprimDeviceSelectTests, Unique)
                 hipFree(d_selected_count_output);
                 hipFree(d_temp_storage);
 
-                if (TestFixture::use_graphs)
-                    test_utils::cleanupGraphHelper(graph, graph_instance);
+                if(TestFixture::use_graphs)
+                {
+                    gHelper.cleanupGraphHelper();
+                }
             }
         }
     }
 
-    if (TestFixture::use_graphs)
+    if(TestFixture::use_graphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
 
 // The operator must be only called, when we have valid element in a block
@@ -690,11 +690,6 @@ void testUniqueGuardedOperator()
                     }
                 }
 
-                hipGraph_t graph;
-                hipGraphExec_t graph_instance;
-                if (UseGraphs)
-                    graph = test_utils::createGraphHelper(stream);
-
                 // temp storage
                 size_t temp_storage_size_bytes;
                 // Get size of d_temp_storage
@@ -712,9 +707,6 @@ void testUniqueGuardedOperator()
                     )
                 );
 
-                if (UseGraphs)
-                    graph_instance = graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
-
                 HIP_CHECK(hipDeviceSynchronize());
 
                 // temp_storage_size_bytes must be >0
@@ -725,8 +717,11 @@ void testUniqueGuardedOperator()
                 HIP_CHECK(hipMalloc(&d_temp_storage, temp_storage_size_bytes));
                 HIP_CHECK(hipDeviceSynchronize());
 
-                if (UseGraphs)
-                    test_utils::resetGraphHelper(graph, graph_instance, stream);
+                test_utils::GraphHelper gHelper;;
+                if(UseGraphs)
+                {
+                    gHelper.startStreamCapture(stream);
+                }
 
                 // Run
                 HIP_CHECK(
@@ -743,8 +738,11 @@ void testUniqueGuardedOperator()
                     )
                 );
 
-                if (UseGraphs)
-                    graph_instance = graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
+                
+                if(UseGraphs)
+                {
+                    gHelper.createAndLaunchGraph(stream, true, false);
+                }
 
                 HIP_CHECK(hipDeviceSynchronize());
 
@@ -778,14 +776,18 @@ void testUniqueGuardedOperator()
                 hipFree(d_selected_count_output);
                 hipFree(d_temp_storage);
 
-                if (UseGraphs)
-                    test_utils::cleanupGraphHelper(graph, graph_instance);
+                if(UseGraphs)
+                {
+                    gHelper.cleanupGraphHelper();
+                }
             }
         }
     }
 
-    if (UseGraphs)
+    if(UseGraphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
 
 TEST(RocprimDeviceSelectTests, UniqueGuardedOperator)
@@ -938,11 +940,6 @@ TYPED_TEST(RocprimDeviceUniqueByKeyTests, UniqueByKey)
                     }
                 }
 
-                hipGraph_t graph;
-                hipGraphExec_t graph_instance;
-                if (TestFixture::use_graphs)
-                    graph = test_utils::createGraphHelper(stream);
-
                 // temp storage
                 size_t temp_storage_size_bytes;
                 // Get size of d_temp_storage
@@ -962,9 +959,6 @@ TYPED_TEST(RocprimDeviceUniqueByKeyTests, UniqueByKey)
                     )
                 );
 
-                if (TestFixture::use_graphs)
-                    graph_instance = graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
-
                 HIP_CHECK(hipDeviceSynchronize());
 
                 // temp_storage_size_bytes must be >0
@@ -975,9 +969,12 @@ TYPED_TEST(RocprimDeviceUniqueByKeyTests, UniqueByKey)
                 HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
                 HIP_CHECK(hipDeviceSynchronize());
 
-                if (TestFixture::use_graphs)
-                    test_utils::resetGraphHelper(graph, graph_instance, stream);
-                
+                test_utils::GraphHelper gHelper;;
+                if(TestFixture::use_graphs)
+                {
+                    gHelper.startStreamCapture(stream);
+                }
+
                 // Run
                 HIP_CHECK(
                     rocprim::unique_by_key(
@@ -995,9 +992,12 @@ TYPED_TEST(RocprimDeviceUniqueByKeyTests, UniqueByKey)
                     )
                 );
 
-                if (TestFixture::use_graphs)
-                    graph_instance = graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
                 
+                if(TestFixture::use_graphs)
+                {
+                    gHelper.createAndLaunchGraph(stream, true, false);
+                }
+
                 HIP_CHECK(hipDeviceSynchronize());
 
                 // Check if number of selected value is as expected
@@ -1040,14 +1040,212 @@ TYPED_TEST(RocprimDeviceUniqueByKeyTests, UniqueByKey)
                 hipFree(d_selected_count_output);
                 hipFree(d_temp_storage);
 
-                if (TestFixture::use_graphs)
-                    test_utils::cleanupGraphHelper(graph, graph_instance);
+                if(TestFixture::use_graphs)
+                {
+                    gHelper.cleanupGraphHelper();
+                }
             }
         }
     }
 
-    if (TestFixture::use_graphs)
+    if(TestFixture::use_graphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
+}
+
+TYPED_TEST(RocprimDeviceUniqueByKeyTests, UniqueByKeyAlias)
+{
+    int device_id = test_common_utils::obtain_device_from_ctest();
+    SCOPED_TRACE(testing::Message() << "with device_id = " << device_id);
+    HIP_CHECK(hipSetDevice(device_id));
+
+    // This test checks correctness of in-place unique_by_key (so input keys and values iterators
+    // are passed as output iterators as well)
+    using key_type          = typename TestFixture::key_type;
+    using value_type        = typename TestFixture::value_type;
+    using output_key_type   = key_type;
+    using output_value_type = value_type;
+
+    using op_type = rocprim::equal_to<key_type>;
+
+    using scan_op_type                          = rocprim::plus<key_type>;
+    static constexpr bool use_identity_iterator = TestFixture::use_identity_iterator;
+    const bool            debug_synchronous     = TestFixture::debug_synchronous;
+
+    hipStream_t stream = 0; // default stream
+    if(TestFixture::use_graphs)
+    {
+        // Default stream does not support hipGraph stream capture, so create one
+        HIP_CHECK(hipStreamCreateWithFlags(&stream, hipStreamNonBlocking));
+    }
+
+    for(size_t seed_index = 0; seed_index < random_seeds_count + seed_size; seed_index++)
+    {
+        unsigned int seed_value
+            = seed_index < random_seeds_count ? rand() : seeds[seed_index - random_seeds_count];
+        SCOPED_TRACE(testing::Message() << "with seed = " << seed_value);
+
+        const auto probabilities = get_discontinuity_probabilities();
+        for(auto size : test_utils::get_sizes(seed_value))
+        {
+            SCOPED_TRACE(testing::Message() << "with size = " << size);
+            for(auto p : probabilities)
+            {
+                SCOPED_TRACE(testing::Message() << "with p = " << p);
+
+                // Generate data
+                std::vector<key_type> input_keys(size);
+                {
+                    std::vector<key_type> input01
+                        = test_utils::get_random_data01<key_type>(size, p, seed_value);
+                    std::partial_sum(input01.begin(),
+                                     input01.end(),
+                                     input_keys.begin(),
+                                     scan_op_type());
+                }
+                const auto input_values
+                    = test_utils::get_random_data<value_type>(size, -1000, 1000, seed_value);
+
+                // Allocate and copy to device
+                key_type*     d_keys_input;
+                value_type*   d_values_input;
+                unsigned int* d_selected_count_output;
+                HIP_CHECK(
+                    test_common_utils::hipMallocHelper(&d_keys_input,
+                                                       input_keys.size() * sizeof(input_keys[0])));
+                HIP_CHECK(test_common_utils::hipMallocHelper(&d_values_input,
+                                                             input_values.size()
+                                                                 * sizeof(input_values[0])));
+                HIP_CHECK(test_common_utils::hipMallocHelper(&d_selected_count_output,
+                                                             sizeof(unsigned int)));
+                HIP_CHECK(hipMemcpy(d_keys_input,
+                                    input_keys.data(),
+                                    input_keys.size() * sizeof(input_keys[0]),
+                                    hipMemcpyHostToDevice));
+                HIP_CHECK(hipMemcpy(d_values_input,
+                                    input_values.data(),
+                                    input_values.size() * sizeof(input_values[0]),
+                                    hipMemcpyHostToDevice));
+                HIP_CHECK(hipDeviceSynchronize());
+
+                // Calculate expected results on host
+                std::vector<output_key_type>   expected_keys;
+                std::vector<output_value_type> expected_values;
+                expected_keys.reserve(input_keys.size());
+                expected_values.reserve(input_values.size());
+                if(size > 0)
+                {
+                    expected_keys.push_back(input_keys[0]);
+                    expected_values.push_back(input_values[0]);
+                    for(size_t i = 1; i < input_keys.size(); i++)
+                    {
+                        if(!op_type()(input_keys[i - 1], input_keys[i]))
+                        {
+                            expected_keys.push_back(input_keys[i]);
+                            expected_values.push_back(input_values[i]);
+                        }
+                    }
+                }
+
+                // temp storage
+                size_t temp_storage_size_bytes;
+                // Get size of d_temp_storage
+                HIP_CHECK(rocprim::unique_by_key(
+                    nullptr,
+                    temp_storage_size_bytes,
+                    d_keys_input,
+                    d_values_input,
+                    test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_keys_input),
+                    test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_values_input),
+                    d_selected_count_output,
+                    input_keys.size(),
+                    op_type(),
+                    stream,
+                    debug_synchronous));
+
+                HIP_CHECK(hipDeviceSynchronize());
+
+                // temp_storage_size_bytes must be >0
+                ASSERT_GT(temp_storage_size_bytes, 0);
+
+                // allocate temporary storage
+                void* d_temp_storage = nullptr;
+                HIP_CHECK(
+                    test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
+                HIP_CHECK(hipDeviceSynchronize());
+
+                test_utils::GraphHelper gHelper;;
+                if(TestFixture::use_graphs)
+                {
+                    gHelper.startStreamCapture(stream);
+                }
+
+                // Run
+                HIP_CHECK(rocprim::unique_by_key(
+                    d_temp_storage,
+                    temp_storage_size_bytes,
+                    d_keys_input,
+                    d_values_input,
+                    test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_keys_input),
+                    test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_values_input),
+                    d_selected_count_output,
+                    input_keys.size(),
+                    op_type(),
+                    stream,
+                    debug_synchronous));
+
+                
+                if(TestFixture::use_graphs)
+                {
+                    gHelper.createAndLaunchGraph(stream);
+                }
+
+                HIP_CHECK(hipDeviceSynchronize());
+
+                // Check if number of selected value is as expected
+                unsigned int selected_count_output = 0;
+                HIP_CHECK(hipMemcpy(&selected_count_output,
+                                    d_selected_count_output,
+                                    sizeof(unsigned int),
+                                    hipMemcpyDeviceToHost));
+                HIP_CHECK(hipDeviceSynchronize());
+                ASSERT_EQ(selected_count_output, expected_keys.size());
+
+                // Check if outputs are as expected
+                std::vector<output_key_type> output_keys(input_keys.size());
+                HIP_CHECK(hipMemcpy(output_keys.data(),
+                                    d_keys_input,
+                                    output_keys.size() * sizeof(output_keys[0]),
+                                    hipMemcpyDeviceToHost));
+                std::vector<output_value_type> output_values(input_values.size());
+                HIP_CHECK(hipMemcpy(output_values.data(),
+                                    d_values_input,
+                                    output_values.size() * sizeof(output_values[0]),
+                                    hipMemcpyDeviceToHost));
+                HIP_CHECK(hipDeviceSynchronize());
+                ASSERT_NO_FATAL_FAILURE(
+                    test_utils::assert_eq(output_keys, expected_keys, expected_keys.size()));
+                ASSERT_NO_FATAL_FAILURE(
+                    test_utils::assert_eq(output_values, expected_values, expected_values.size()));
+
+                hipFree(d_keys_input);
+                hipFree(d_values_input);
+                hipFree(d_selected_count_output);
+                hipFree(d_temp_storage);
+
+                if(TestFixture::use_graphs)
+                {
+                    gHelper.cleanupGraphHelper();
+                }
+            }
+        }
+    }
+
+    if(TestFixture::use_graphs)
+    {
+        HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
 
 class RocprimDeviceSelectLargeInputTests : public ::testing::TestWithParam<std::pair<unsigned int, bool>> {
@@ -1123,11 +1321,6 @@ TEST_P(RocprimDeviceSelectLargeInputTests, LargeInputFlagged)
         size_t temp_storage_size_bytes;
         void *d_temp_storage = nullptr;
 
-        hipGraph_t graph;
-        hipGraphExec_t graph_instance;
-        if (use_graphs)
-            graph = test_utils::createGraphHelper(stream);
-
         // Get size of d_temp_storage
         HIP_CHECK(
             rocprim::select(
@@ -1143,9 +1336,6 @@ TEST_P(RocprimDeviceSelectLargeInputTests, LargeInputFlagged)
             )
         );
 
-        if (use_graphs)
-            graph_instance = graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
-
         HIP_CHECK(hipDeviceSynchronize());
 
         // temp_storage_size_bytes must be >0
@@ -1155,8 +1345,11 @@ TEST_P(RocprimDeviceSelectLargeInputTests, LargeInputFlagged)
         HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
         HIP_CHECK(hipDeviceSynchronize());
 
-        if (use_graphs)
-            test_utils::resetGraphHelper(graph, graph_instance, stream);
+        test_utils::GraphHelper gHelper;;
+        if(use_graphs)
+        {
+            gHelper.startStreamCapture(stream);
+        }
 
         // Run
         HIP_CHECK(
@@ -1173,8 +1366,11 @@ TEST_P(RocprimDeviceSelectLargeInputTests, LargeInputFlagged)
             )
         );
 
-        if (use_graphs)
-            graph_instance = graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, false);
+        
+        if(use_graphs)
+        {
+            gHelper.createAndLaunchGraph(stream, true, false);
+        }
 
         HIP_CHECK(hipDeviceSynchronize());
 
@@ -1205,8 +1401,10 @@ TEST_P(RocprimDeviceSelectLargeInputTests, LargeInputFlagged)
         hipFree(d_selected_count_output);
         hipFree(d_temp_storage);
 
-        if (use_graphs)
-            test_utils::cleanupGraphHelper(graph, graph_instance);
+        if(use_graphs)
+        {
+            gHelper.cleanupGraphHelper();
+        }
     }
 
     if (use_graphs)
@@ -1254,11 +1452,6 @@ TEST_P(RocprimDeviceSelectLargeInputTests, LargeInputUnique)
         HIP_CHECK(test_common_utils::hipMallocHelper(&d_unique_count_output,
                                                      sizeof(*d_unique_count_output)));
 
-        hipGraph_t graph;
-        hipGraphExec_t graph_instance;
-        if (use_graphs)
-            graph = test_utils::createGraphHelper(stream);
-
         size_t temp_storage_size_bytes{};
         void*  d_temp_storage{};
         HIP_CHECK(rocprim::unique(d_temp_storage,
@@ -1271,14 +1464,14 @@ TEST_P(RocprimDeviceSelectLargeInputTests, LargeInputUnique)
                                   stream,
                                   debug_synchronous));
 
-        if (use_graphs)
-            graph_instance = graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-
         ASSERT_GT(temp_storage_size_bytes, 0);
         HIP_CHECK(test_common_utils::hipMallocHelper(&d_temp_storage, temp_storage_size_bytes));
 
-        if (use_graphs)
-            test_utils::resetGraphHelper(graph, graph_instance, stream);
+        test_utils::GraphHelper gHelper;;
+        if(use_graphs)
+        {
+            gHelper.startStreamCapture(stream);
+        }
 
         HIP_CHECK(rocprim::unique(d_temp_storage,
                                   temp_storage_size_bytes,
@@ -1290,8 +1483,11 @@ TEST_P(RocprimDeviceSelectLargeInputTests, LargeInputUnique)
                                   stream,
                                   debug_synchronous));
 
-        if (use_graphs)
-            graph_instance = graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
+        
+        if(use_graphs)
+        {
+            gHelper.createAndLaunchGraph(stream);
+        }
 
         size_t unique_count_output{};
         HIP_CHECK(hipMemcpyWithStream(&unique_count_output,
@@ -1316,7 +1512,7 @@ TEST_P(RocprimDeviceSelectLargeInputTests, LargeInputUnique)
         HIP_CHECK(hipFree(d_temp_storage));
 
         if (use_graphs)
-            test_utils::cleanupGraphHelper(graph, graph_instance);
+            gHelper.cleanupGraphHelper();
     }
 
     if (use_graphs)
