@@ -1,4 +1,4 @@
-// Copyright (c) 2021-2023 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2021-2024 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -31,6 +31,10 @@
 
 // Std::memcpy and std::memcmp
 #include <cstring>
+#include <iterator>
+#include <ostream>
+#include <sstream>
+#include <type_traits>
 #include <vector>
 
 namespace test_utils {
@@ -39,7 +43,7 @@ namespace test_utils {
 template<class T>
 bool inline bit_equal(const T& a, const T& b)
 {
-    return std::memcmp(&a,  &b, sizeof(T))==0;
+    return std::memcmp(&a, &b, sizeof(T)) == 0;
 }
 
 /// Checks if `vector<T> result` matches `vector<T> expected`.
@@ -50,58 +54,115 @@ bool inline bit_equal(const T& a, const T& b)
 /// \param expected
 /// \param max_length
 template<class T>
-void assert_eq(const std::vector<T>& result, const std::vector<T>& expected, const size_t max_length = SIZE_MAX)
+void assert_eq(const std::vector<T>& result,
+               const std::vector<T>& expected,
+               const size_t          max_length = SIZE_MAX)
 {
-    if(max_length == SIZE_MAX || max_length > expected.size()) ASSERT_EQ(result.size(), expected.size());
+    if(max_length == SIZE_MAX || max_length > expected.size())
+        ASSERT_EQ(result.size(), expected.size());
     for(size_t i = 0; i < std::min(result.size(), max_length); i++)
     {
-        if(bit_equal(result[i], expected[i])) continue; // Check bitwise equality for +NaN, -NaN, +0.0, -0.0, +inf, -inf.
+        if(bit_equal(result[i], expected[i]))
+            continue; // Check bitwise equality for +NaN, -NaN, +0.0, -0.0, +inf, -inf.
+        ASSERT_EQ(result[i], expected[i]) << "where index = " << i;
+    }
+}
+
+template<class T>
+void assert_eq(const std::vector<custom_test_type<T>>& result,
+               const std::vector<custom_test_type<T>>& expected,
+               const size_t                            max_length = SIZE_MAX)
+{
+    if(max_length == SIZE_MAX || max_length > expected.size())
+    {
+        ASSERT_EQ(result.size(), expected.size());
+    }
+    for(size_t i = 0; i < std::min(result.size(), max_length); i++)
+    {
+        if(bit_equal(result[i].x, expected[i].x) && bit_equal(result[i].y, expected[i].y))
+            continue; // Check bitwise equality for +NaN, -NaN, +0.0, -0.0, +inf, -inf.
         ASSERT_EQ(result[i], expected[i]) << "where index = " << i;
     }
 }
 
 template<>
-inline void assert_eq<rocprim::half>(const std::vector<rocprim::half>& result, const std::vector<rocprim::half>& expected, const size_t max_length)
+inline void assert_eq<rocprim::half>(const std::vector<rocprim::half>& result,
+                                     const std::vector<rocprim::half>& expected,
+                                     const size_t                      max_length)
 {
-    if(max_length == SIZE_MAX || max_length > expected.size()) ASSERT_EQ(result.size(), expected.size());
+    if(max_length == SIZE_MAX || max_length > expected.size())
+        ASSERT_EQ(result.size(), expected.size());
     for(size_t i = 0; i < std::min(result.size(), max_length); i++)
     {
-        if(bit_equal(result[i], expected[i])) continue; // Check bitwise equality for +NaN, -NaN, +0.0, -0.0, +inf, -inf.
+        if(bit_equal(result[i], expected[i]))
+            continue; // Check bitwise equality for +NaN, -NaN, +0.0, -0.0, +inf, -inf.
         ASSERT_EQ(half_to_native(result[i]), half_to_native(expected[i])) << "where index = " << i;
     }
 }
 
 template<>
-inline void assert_eq<rocprim::bfloat16>(const std::vector<rocprim::bfloat16>& result, const std::vector<rocprim::bfloat16>& expected, const size_t max_length)
+inline void assert_eq<rocprim::bfloat16>(const std::vector<rocprim::bfloat16>& result,
+                                         const std::vector<rocprim::bfloat16>& expected,
+                                         const size_t                          max_length)
 {
     if(max_length == SIZE_MAX || max_length > expected.size()) ASSERT_EQ(result.size(), expected.size());
     for(size_t i = 0; i < std::min(result.size(), max_length); i++)
     {
-        if(bit_equal(result[i], expected[i])) continue; // Check bitwise equality for +NaN, -NaN, +0.0, -0.0, +inf, -inf.
-        ASSERT_EQ(bfloat16_to_native(result[i]), bfloat16_to_native(expected[i])) << "where index = " << i;
+        if(bit_equal(result[i], expected[i]))
+            continue; // Check bitwise equality for +NaN, -NaN, +0.0, -0.0, +inf, -inf.
+        ASSERT_EQ(bfloat16_to_native(result[i]), bfloat16_to_native(expected[i]))
+            << "where index = " << i;
     }
 }
 
 template<class T>
 void assert_eq(const T& result, const T& expected)
 {
-    if(bit_equal(result, expected)) return; // Check bitwise equality for +NaN, -NaN, +0.0, -0.0, +inf, -inf.
+    if(bit_equal(result, expected))
+        return; // Check bitwise equality for +NaN, -NaN, +0.0, -0.0, +inf, -inf.
+    ASSERT_EQ(result, expected);
+}
+
+template<class T>
+void assert_eq(const custom_test_type<T>& result, const custom_test_type<T>& expected)
+{
+    if(bit_equal(result.x, expected.x) && bit_equal(result.y, expected.y))
+        return; // Check bitwise equality for +NaN, -NaN, +0.0, -0.0, +inf, -inf.
     ASSERT_EQ(result, expected);
 }
 
 template<>
 inline void assert_eq<rocprim::half>(const rocprim::half& result, const rocprim::half& expected)
 {
-    if(bit_equal(result, expected)) return; // Check bitwise equality for +NaN, -NaN, +0.0, -0.0, +inf, -inf.
+    if(bit_equal(result, expected))
+        return; // Check bitwise equality for +NaN, -NaN, +0.0, -0.0, +inf, -inf.
     ASSERT_EQ(half_to_native(result), half_to_native(expected));
 }
 
 template<>
-inline void assert_eq<rocprim::bfloat16>(const rocprim::bfloat16& result, const rocprim::bfloat16& expected)
+inline void assert_eq<rocprim::bfloat16>(const rocprim::bfloat16& result,
+                                         const rocprim::bfloat16& expected)
 {
-    if(bit_equal(result, expected)) return; // Check bitwise equality for +NaN, -NaN, +0.0, -0.0, +inf, -inf.
+    if(bit_equal(result, expected))
+        return; // Check bitwise equality for +NaN, -NaN, +0.0, -0.0, +inf, -inf.
     ASSERT_EQ(bfloat16_to_native(result), bfloat16_to_native(expected));
 }
+
+template<class ResultIt, class ExpectedIt>
+void assert_eq(ResultIt   result_begin,
+               ResultIt   result_end,
+               ExpectedIt expected_begin,
+               ExpectedIt expected_end)
+{
+    ASSERT_EQ(std::distance(result_begin, result_end), std::distance(expected_begin, expected_end));
+    auto result_it   = result_begin;
+    auto expected_it = expected_begin;
+    for(; result_it != result_end; ++result_it, ++expected_it)
+    {
+        assert_eq(static_cast<typename std::iterator_traits<ResultIt>::value_type>(*result_it),
+                  static_cast<typename std::iterator_traits<ExpectedIt>::value_type>(*expected_it));
+    }
+}
 // end assert_eq
 
 // begin assert_near
@@ -113,7 +174,7 @@ auto assert_near(const std::vector<T>& result, const std::vector<T>& expected, c
     for(size_t i = 0; i < result.size(); i++)
     {
         if(bit_equal(result[i], expected[i])) continue; // Check bitwise equality for +NaN, -NaN, +0.0, -0.0, +inf, -inf.
-        auto diff = std::abs(percent * expected[i]);
+        auto diff = std::abs(percent * std::max(result[i], expected[i]));
         ASSERT_NEAR(result[i], expected[i], diff) << "where index = " << i;
     }
 }
@@ -230,97 +291,84 @@ auto assert_near(const custom_test_type<T>& result, const custom_test_type<T>& e
 
 // End assert_near
 
+#if ROCPRIM_HAS_INT128_SUPPORT
 template<class T>
+auto operator<<(std::ostream& os, const T& value)
+    -> std::enable_if_t<std::is_same<T, __int128_t>::value || std::is_same<T, __uint128_t>::value,
+                        std::ostream&>
+{
+    static const char* charmap = "0123456789";
+
+    std::string result;
+    result.reserve(41); // max. 40 digits possible ( uint64_t has 20) plus sign
+    __uint128_t helper = (value < 0) ? -value : value;
+
+    do
+    {
+        result += charmap[helper % 10];
+        helper /= 10;
+    }
+    while(helper);
+    if(value < 0)
+    {
+        result += "-";
+    }
+    std::reverse(result.begin(), result.end());
+
+    os << result;
+    return os;
+}
+
+#endif
+
+template<typename T>
+testing::AssertionResult
+    bitwise_equal(const char* a_expr, const char* b_expr, const T& a, const T& b)
+{
+    if(bit_equal(a, b))
+    {
+        return testing::AssertionSuccess();
+    }
+
+    // googletest's operator<< doesn't see the above overload for int128_t
+    std::stringstream a_str;
+    std::stringstream b_str;
+    a_str << std::hexfloat << a;
+    b_str << std::hexfloat << b;
+
+    return testing::AssertionFailure()
+           << "Expected strict/bitwise equality of these values: " << std::endl
+           << "     " << a_expr << ": " << std::hexfloat << a_str.str() << std::endl
+           << "     " << b_expr << ": " << std::hexfloat << b_str.str() << std::endl;
+}
+
+#define ASSERT_BITWISE_EQ(a, b) ASSERT_PRED_FORMAT2(bitwise_equal, a, b)
+
+template<typename IterA, typename IterB>
+void assert_bit_eq(IterA result_begin, IterA result_end, IterB expected_begin, IterB expected_end)
+{
+    using value_a_t = typename std::iterator_traits<IterA>::value_type;
+    using value_b_t = typename std::iterator_traits<IterB>::value_type;
+
+    ASSERT_EQ(std::distance(result_begin, result_end), std::distance(expected_begin, expected_end));
+    auto result_it   = result_begin;
+    auto expected_it = expected_begin;
+    for(size_t index = 0; result_it != result_end; ++result_it, ++expected_it, ++index)
+    {
+        // The cast is needed, because the argument can be an std::vector<bool> iterator, which's operator*
+        // returns a proxy object that must be converted to bool
+        const auto result   = static_cast<value_a_t>(*result_it);
+        const auto expected = static_cast<value_b_t>(*expected_it);
+
+        ASSERT_BITWISE_EQ(result, expected) << "where index = " << index;
+    }
+}
+
+template<typename T>
 void assert_bit_eq(const std::vector<T>& result, const std::vector<T>& expected)
 {
-    ASSERT_EQ(result.size(), expected.size());
-    for(size_t i = 0; i < result.size(); i++)
-    {
-        if(!bit_equal(result[i], expected[i]))
-        {
-            FAIL() << "Expected strict/bitwise equality of these values: " << std::endl
-                   << "     result[i]: " << result[i] << std::endl
-                   << "     expected[i]: " << expected[i] << std::endl
-                   << "where index = " << i;
-        }
-    }
+    assert_bit_eq(result.begin(), result.end(), expected.begin(), expected.end());
 }
-
-#if defined(__GNUC__) || defined(__clang__)
-inline void assert_bit_eq(const std::vector<__int128_t>& result,
-                          const std::vector<__int128_t>& expected)
-{
-    ASSERT_EQ(result.size(), expected.size());
-
-    auto to_string = [](__int128_t value)
-    {
-        static const char* charmap = "0123456789";
-
-        std::string result;
-        result.reserve(41); // max. 40 digits possible ( uint64_t has 20) plus sign
-        __uint128_t helper = (value < 0) ? -value : value;
-
-        do
-        {
-            result += charmap[helper % 10];
-            helper /= 10;
-        }
-        while(helper);
-        if(value < 0)
-        {
-            result += "-";
-        }
-        std::reverse(result.begin(), result.end());
-        return result;
-    };
-
-    for(size_t i = 0; i < result.size(); i++)
-    {
-        if(!bit_equal(result[i], expected[i]))
-        {
-            FAIL() << "Expected strict/bitwise equality of these values: " << std::endl
-                   << "     result[i]: " << to_string(result[i]) << std::endl
-                   << "     expected[i]: " << to_string(expected[i]) << std::endl
-                   << "where index = " << i;
-        }
-    }
-}
-
-inline void assert_bit_eq(const std::vector<__uint128_t>& result,
-                          const std::vector<__uint128_t>& expected)
-{
-    ASSERT_EQ(result.size(), expected.size());
-
-    auto to_string = [](__uint128_t value)
-    {
-        static const char* charmap = "0123456789";
-
-        std::string result;
-        result.reserve(40); // max. 40 digits possible ( uint64_t has 20)
-        __uint128_t helper = value;
-
-        do
-        {
-            result += charmap[helper % 10];
-            helper /= 10;
-        }
-        while(helper);
-        std::reverse(result.begin(), result.end());
-        return result;
-    };
-
-    for(size_t i = 0; i < result.size(); i++)
-    {
-        if(!bit_equal(result[i], expected[i]))
-        {
-            FAIL() << "Expected strict/bitwise equality of these values: " << std::endl
-                   << "     result[i]: " << to_string(result[i]) << std::endl
-                   << "     expected[i]: " << to_string(expected[i]) << std::endl
-                   << "where index = " << i;
-        }
-    }
-}
-#endif
 }
 
 #endif //ROCPRIM_TEST_UTILS_ASSERTIONS_HPP
